import caffi.ca as ca


def setup_module(module):
    global pvs, vals, gid
    pvs = {}
    vals = {}

    # create channels
    for name in ['cawaves', 'cawaveh', 'cawavef', 'cawavec', 'cawavel', 'cawave']:
        status, chid = ca.create_channel(name)
        assert status == ca.ECA.NORMAL
        pvs[name] = chid

    # wait for connections
    status = ca.pend_io(10)
    assert status == ca.ECA.NORMAL

    # create synchronous group
    status, gid = ca.sg_create()
    assert status == ca.ECA.NORMAL


def test_put_get():
    # put
    for name, chid in pvs.items():
        if ca.field_type(chid) == ca.DBF.STRING:
            ca.sg_put(gid, chid, ['1', '2', '3', '4'])
        else:
            ca.sg_put(gid, chid, [1, 2, 3, 4])

    ca.flush_io()

    status = ca.sg_block(gid, 3)
    assert status == ca.ECA.NORMAL

    # get
    for name, chid in pvs.items():
        status, dbrvalue = ca.sg_get(gid, chid, count=4)
        assert status == ca.ECA.NORMAL
        vals[name] = dbrvalue

    ca.flush_io()

    # wait for get completion
    status = ca.sg_block(gid, 3)
    assert status == ca.ECA.NORMAL

    # compare read value with written values
    for name, dbrvalue in vals.items():
        chid = pvs[name]
        element_count = ca.element_count(chid)
        value = dbrvalue.get()
        if ca.field_type(chid) == ca.DBF.STRING:
            assert value == ['1', '2', '3', '4'][:element_count]
        else:
            assert value == [1, 2, 3, 4][:element_count]


def teardown_module():
    # delete synchronous group
    ca.sg_delete(gid)

    # close channels
    for name, chid in pvs.items():
        ca.clear_channel(chid)

    ca.flush_io()

    pvs.clear()
    vals.clear()



"""
The is the low level access to EPICS channel access library. It maps the corresponding C API to Python functions.
Even though as same as possible, there are subtle differences:

  - The `ca_` prefix of the C function name has been removed, e.g. C function `ca_create_channel` is now
    Python function :func:`create_channel`.

  - The order of the argument might have been altered to allow default arguments, thus more pythonic.
    Take C function `array_put_callback` for example,
    ::

        int ca_array_put_callback
        (
            chtype                 type,
            unsigned long          count,
            chid                   chanId,
            const void *           pValue,
            caEventCallBackFunc *  pFunc,
            void *                 pArg
        );

    Its Python counterpart,
    ::

        put(chid, value, chtype=None, count=None, callback=None)

    Only two arguments are mandatory, which are the channel identifier and the value to write.
    The others are made optional and have reasonable defaults.

  - In C API the callback function handler has the following signature,
    ::

        void (*)(struct xxx_handler_args)

    In Python counterpart, the callback signature is
    ::

        def callback(epicsArgs):

    *epicsArgs* is a dict converted from `xxx_handler_args`.

  - C functions normally return status code to indicate success or failure. The Python counterparts follow
    this but have exceptions:

    - The get functions in C require a user supplied memory pointer passed as argument. In Python the function
      :func:`get` and :func:`sg_get` returns a tuple of form (:class:`caffi.ca.ECA`, :class:`caffi.ca.DBRValue`)

      The first item is the status code, which must be :data:`caffi.ca.ECA.NORMAL`, before using the second item.

      The second item holds the reference to the allocated memory. And the memory is not stable until a subsequent
      call of :func:`pend_io` returns :data:`caffi.ca.ECA.NORMAL`. Then the value can be retrieved by calling
      :meth:`caffi.ca.DBRValue.get`.


    - All the creation functions, :func:`create_channel`, :func:`create_subscription` and :func:`sg_create`
      return a tuple of the form (:class:`caffi.ca.ECA`, *object identifier*).
      The object identifier can only be used if the first item is :data:`caffi.ca.ECA.NORMAL`.

  - In C the following macros definition are also accessible as :class:`enum.IntEnum` type:

    ======================  ============
    C Macros                Python Enum
    ======================  ============
    *ECA_XXX*               :class:`caffi.ca.ECA`
    *DBE_XXX*               :class:`caffi.ca.DBE`
    *DBF_XXX*               :class:`caffi.ca.DBF`
    *DBR_XXX*               :class:`caffi.ca.DBR`
    *CA_OP_XXX*             :class:`caffi.ca.CA_OP`
    *cs_xxx*                :class:`caffi.ca.ChannelState`
    *XXX_ALARM* (severity)  :class:`caffi.ca.AlarmSeverity`
    *XXX_ALARM* (status)    :class:`caffi.ca.AlarmCondition`
    ======================  ============

    This makes it convenient when interactively examine the code value. e.g.
    ::

        >>> ca.create_context(True)
        <ECA.NORMAL: 1>
        >>> status, chid = ca.create_channel('catest')
        >>> ca.pend_io(2)
        <ECA.NORMAL: 1>
        >>> ca.name(chid)
        'catest'
        >>> ca.host_name(chid)
        'localhost:5064'
        >>> ca.field_type(chid)
        <DBF.DOUBLE: 6>
        >>> ca.element_count(chid)
        1
        >>> ca.read_access(chid)
        True
        >>> ca.write_access(chid)
        True
        >>> ca.put(chid, 10)
        <ECA.NORMAL: 1>
        >>> ca.flush_io()
        <ECA.NORMAL: 1>
        >>> status, dbrvalue = ca.get(chid)
        >>> ca.pend_io(2)
        <ECA.NORMAL: 1>
        >>> dbrvalue.get()
        10.0

"""
from __future__ import (print_function, absolute_import)
import collections
import numbers
import sys

from .compat import *
from ._ca import *
from .constants import *
from .dbr import *
from .macros import *

__all__ = ['create_context', 'current_context', 'attach_context', 'detach_context', 'destroy_context', 'show_context',
           'add_exception_event', 'replace_access_rights_event', 'change_connection_event',
           'create_channel', 'clear_channel', 'get', 'put', 'create_subscription', 'clear_subscription',
           'field_type', 'element_count', 'name', 'state', 'host_name', 'read_access', 'write_access',
           'pend_event', 'pend_io', 'poll', 'pend', 'flush_io', 'test_io', 'message',
           'sg_create', 'sg_delete', 'sg_get', 'sg_put', 'sg_reset', 'sg_block', 'sg_test', 'version']

# globals
__channels = {}
__exception_callback = {}

DBR_TYPE_STRING = {
    DBR.STRING:   'dbr_string_t',
    DBR.INT:      'dbr_int_t',
    DBR.FLOAT:    'dbr_float_t',
    DBR.CHAR:     'dbr_char_t',
    DBR.ENUM:     'dbr_enum_t',
    DBR.LONG:     'dbr_long_t',
    DBR.DOUBLE:   'dbr_double_t',
    DBR.PUT_ACKT: 'dbr_put_ackt_t',
    DBR.PUT_ACKS: 'dbr_put_acks_t',
}


@ffi.callback('void(*)(struct exception_handler_args)')
def _exception_callback(arg):
    if arg.pFile == ffi.NULL:
        file = ''
    else:
        file = to_string(ffi.string(arg.pFile))

    epics_arg = {
        'chid':   arg.chid,
        'type':   DBR(arg.type),
        'count':  arg.count,
        'addr':   arg.addr,
        'stat':   ECA(arg.stat),
        'op':     CA_OP(arg.op),
        'ctx':    to_string(ffi.string(arg.ctx)),
        'file':   file,
        'lineNo': arg.lineNo
    }

    user_callback = ffi.from_handle(arg.usr)
    if callable(user_callback):
        user_callback(epics_arg)


def add_exception_event(callback=None):
    """
    Replace the currently installed CA context global exception handler call back.

    :param callback:    User callback function to be executed when an exceptions occur.
                        Passing None causes the default exception handler to be reinstalled.
                        The argument is a dict including the following fields:

                        ======  =============
                        field   value
                        ======  =============
                        chid    channel identifier (may be NULL)
                        type    type requested
                        count   count requested
                        addr    user's address to write results of CA_OP.GET (may be NULL)
                        stat    status code, :class:`caffi.ca.ECA`
                        op      operation, :class:`caffi.ca.CA_OP`
                        ctx     a character string containing context info
                        file    source file name (may be empty)
                        lineNo  source file line number (may be zero)
                        ======  =============

    :type callback:     callable, None
    :return: ECA.NORMAL - Normal successful completion

    When an error occurs in the server asynchronous to the clients thread then information about this type of error
    is passed from the server to the client in an exception message. When the client receives this exception message
    an exception handler callback is called.

    The default exception handler prints a diagnostic message on the client's standard out and
    terminates execution if the error condition is severe.

    .. note::   Certain fields are not applicable in the context of some error messages.
                For instance, a failed get will supply the address in the client task
                where the returned value was requested to be written.
                For other failed operations the value of the *addr* field should not be used.

    """
    if callable(callback):
        callback_handle = ffi.new_handle(callback)
        status = libca.ca_add_exception_event(_exception_callback, callback_handle)
    else:
        callback_handle = None
        status = libca.ca_add_exception_event(ffi.NULL, ffi.NULL)

    # keep a reference to the returned pointer of (callback, args),
    # otherwise it will be garbage collected
    context = libca.ca_current_context()
    __exception_callback[context] = callback_handle

    return ECA(status)


def create_context(preemptive_callback=True):
    """
    :param bool preemptive_callback: enable preemptive callback
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.ALLOCMEM - Failed, unable to allocate space in pool
        - ECA.NOTTHREADED - Current thread is already a member of a non-preemptive callback CA context
                                    (possibly created implicitly)

    This function, or :func:`attach_context`, should be called once from each thread prior to
    making any of the other Channel Access calls. If one of the above is not called before
    making other CA calls then a non-preemptive context is created by default,
    and future attempts to create a preemptive context for the current threads will fail.

    If preemptive callback is disabled then additional threads are not allowed to
    join the CA context using :func:`attach_context` because allowing other threads to join implies
    that CA callbacks will be called preemptively from more than one thread.

    """
    if preemptive_callback:
        status = libca.ca_context_create(libca.ca_enable_preemptive_callback)
    else:
        status = libca.ca_context_create(libca.ca_disable_preemptive_callback)

    return ECA(status)


def destroy_context():
    """
    Shut down the calling thread's channel access client context and free any resources allocated.
    Detach the calling thread from any CA client context.

    Any user-created threads that have attached themselves to the CA context must stop using it
    prior to its being destroyed.

    On many OS that execute programs in a process based environment the resources used by the client library
    such as sockets and allocated memory are automatically released by the system when the process exits and
    :func:`destroy_context` hasn't been called, but on light weight systems such as vxWorks or RTEMS
    no cleanup occurs unless the application calls :func:`destroy_context`.
    """
    context = libca.ca_current_context()
    if context != ffi.NULL and context in __exception_callback:
        del __exception_callback[context]

    libca.ca_context_destroy()


def attach_context(context):
    """
    :param cdata context: The CA context to join with.
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.NOTTHREADED - Context is not preemptive so cannot be joined
        - ECA.ISATTACHED - Thread already attached to a CA context

    The calling thread becomes a member of the specified CA context.
    If context is non preemptive, then additional threads are not allowed to join the CA context
    because allowing other threads to join implies that CA callbacks will be called preemptively
    from more than one thread.
    """
    status = libca.ca_attach_context(context)
    return ECA(status)


def detach_context():
    """
    Detach from any CA context currently attached to the calling thread.

    This does not cleanup or shutdown any currently attached CA context.
    """
    libca.ca_detach_context()


def current_context():
    """
    :return: The current thread's CA context. If none then None is returned.
    """
    context = libca.ca_current_context()
    if context == ffi.NULL:
        context = None
    return context


def show_context(context=None, level=0):
    """
    Prints information about the client context including, at higher interest levels, status for each channel.

    :param context: The CA context to examine. Default is the calling threads CA context.
    :param level:   The interest level. Increasing level produces increasing detail.
    :type context:  cdata, None
    :type level:    int
    """
    if context is None:
        libca.ca_client_status(level)
    else:
        libca.ca_context_status(context, level)


@ffi.callback('void(*)(struct connection_handler_args)')
def _connect_callback(arg):
    epics_arg = {
        'chid': arg.chid,
        'op':   CA_OP(arg.op)
    }

    # If chid is not in cache, it well indicates
    # that the python object has been garbage collected.
    # Then don't try to call from_handle, that is undefined and may crash.
    if arg.chid not in __channels:
        return

    user_callback = __channels[arg.chid]['connection_callback']

    if callable(user_callback):
        user_callback(epics_arg)


def create_channel(name, callback=None, priority=CA_PRIORITY.DEFAULT):
    """
    This function creates a CA channel.

    :param name:        Process variable name string.
    :param callback:    Optional user's call back function to be run when the connection state changes.
                        Casual users of channel access may decide to leave it None if they do not need to
                        have a call back function run in response to each connection state change event.
                        The callback argument is a dict including the following fields:

                        =====   =============
                        field   value
                        =====   =============
                        chid    channel identifier
                        op      - CA_OP.CONN_UP - connected
                                - CA_OP.CONN_DOWN - disconnected
                        =====   =============

    :param priority:    The priority level for dispatch within the server or network,
                        with 0 specifying the lowest dispatch priority and 99 the highest.
                        This parameter currently does not impact dispatch priorities within the client,
                        but this might change in the future.
                        The abstract priority range specified is mapped into an operating system specific
                        range of priorities within the server.
                        This parameter is ignored if the server is running on a network or operating system
                        that does not have native support for prioritized delivery or execution respectively.
                        Specifying many different priorities within the same program can increase resource consumption
                        in the client and the server because an independent virtual circuit, and associated
                        data structures, is created for each priority that is used on a particular server.
    :type name:         str
    :type callback:     callable, None
    :type priority:     int, :class:`caffi.ca.CA_PRIORITY`
    :return:            (status code, channel identifier)
    :rtype:             (:class:`caffi.ca.ECA`, cdata)

    The CA client library will attempt to establish and maintain a virtual circuit between the caller's application
    and a named process variable in a CA server. Each call to ca_create_channel allocates resources
    in the CA client library and potentially also a CA server.
    The function :func:`clear_channel` is used to release these resources.

    If successful, the routine returns a channel identifier. This identifier can be used with any channel access call
    that operates on a channel.

    The circuit may be initially connected or disconnected depending on the state of the network
    and the location of the channel. A channel will only enter a connected state
    after the server's address is determined, and only if channel access successfully establishes a virtual circuit
    through the network to the server. Channel access routines that send a request to a server
    will return :data:`caffi.ca.ECA.DISCONNCHID` if the channel is currently disconnected.

    There are two ways to obtain asynchronous notification when a channel enters a connected state.

    - The first and simplest method requires that you call ca_pend_io(), and wait for successful completion,
      prior to using a channel that was created without specifying call back function.

    - The second method requires that you register a connection handler
      by supplying a valid connection callback function pointer. This connection handler is called whenever
      the connection state of the channel changes. If you have installed a connection handler
      then :func:`pend_io` will not block waiting for the channel to enter a connected state.

    The function :func:`state` can be used to test the connection state of a channel.
    Valid connections may be isolated from invalid ones with this function :func:`pend_io` times out.

    Due to the inherently transient nature of network connections the order of connection call backs
    relative to the order that :func:`create_channel` calls are made by the application
    can't be guaranteed, and application programs may need to be prepared for a connected channel to
    enter a disconnected state at any time.

    """
    name = to_bytes(name)

    pchid = ffi.new('chid *')

    if callable(callback):
        connect_callback = _connect_callback
    else:
        connect_callback = ffi.NULL

    status = libca.ca_create_channel(name, connect_callback, ffi.NULL, priority, pchid)
    if status != ECA_NORMAL:
        return ECA(status), None

    chid = pchid[0]
    __channels[chid] = {'callbacks': set(), 'monitors': {}}

    if callable(callback):
        __channels[chid]['connection_callback'] = callback
    else:
        __channels[chid]['connection_callback'] = None

    return ECA(status), chid


def change_connection_event(chid, callback=None):
    """
    Change the connection event callback function.

    :param chid:        Channel identifier
    :param callback:    User's call back function to be run when the connection state changes. The callback receives
                        the same argument as :func:`create_channel`. This will replace the previous connection callback
                        function. If an invalid *callback* is given, no connection callback is used.
    :return:
        - ECA.NORMAL - Normal successful completion
    """
    if chid not in __channels:
        return ECA.BADCHID

    if callable(callback):
        __channels[chid]['connection_callback'] = callback
        status = libca.ca_change_connection_event(chid, _connect_callback)
    else:
        __channels[chid]['connection_callback'] = None
        status = libca.ca_change_connection_event(chid, ffi.NULL)

    # store the reference so it won't be garbage collected
    return ECA(status)


@ffi.callback('void(*)(struct access_rights_handler_args)')
def _access_rights_callback(arg):
    epics_arg = {
        'chid': arg.chid,
        'read_access': arg.ar.access & 0x01 == 0x01,
        'write_access': arg.ar.access & 0x02 == 0x02
    }
    # If chid is not in cache, it well indicates
    # that the python object has been garbage collected.
    # Then don't try to call from_handle, that is undefined and may crash.
    if arg.chid not in __channels:
        return

    callback = __channels[arg.chid]['access_rights_callback']
    if callable(callback):
        callback(epics_arg)


def replace_access_rights_event(chid, callback=None):
    """
    Install or replace the access rights state change callback handler for the specified channel.

    :param chid:     The channel identifier
    :param callback: User supplied call back function. Passing None uninstalls the current handler.
                     The callback argument is a dict including the following fields:

                     ============   =============
                     field          value
                     ============   =============
                     chid           channel identifier
                     read_access    True if with read access rights
                     write_access   True if with write access rights
                     ============   =============

    :type chid:      cdata
    :type callback:  callable, None
    :return:
        - ECA.NORMAL - Normal successful completion

    The callback handler is called in the following situations.

        - whenever CA connects the channel immediately before the channel's connection handler is called
        - whenever CA disconnects the channel immediately after the channel's disconnect call back is called
        - once immediately after installation if the channel is connected.
        - whenever the access rights state of a connected channel changes

    When a channel is created no access rights handler is installed.
    """
    if chid not in __channels:
        return ECA.BADCHID

    if callable(callback):
        __channels[chid]['access_rights_callback'] = callback
        status = libca.ca_replace_access_rights_event(chid, _access_rights_callback)
    else:
        __channels[chid]['access_rights_callback'] = None
        status = libca.ca_replace_access_rights_event(chid, ffi.NULL)

    return ECA(status)


@ffi.callback('void(struct event_handler_args)')
def _get_callback(arg):
    # If chid or the callback object is not in cache, it well indicates
    # that the python object has been garbage collected.
    # Then don't try to call from_handle, that is undefined and may crash.
    if arg.chid not in __channels or arg.usr not in __channels[arg.chid]['callbacks']:
        return

    user_callback, use_numpy = ffi.from_handle(arg.usr)
    __channels[arg.chid]['callbacks'].remove(arg.usr)

    epics_arg = {
        'chid':   arg.chid,
        'type':   DBR(arg.type),
        'count':  arg.count,
        'status': ECA(arg.status),
        'value':  format_dbr(arg.type, arg.count, arg.dbr, use_numpy)
    }
    if callable(user_callback):
        user_callback(epics_arg)


def get(chid, chtype=None, count=None, callback=None, use_numpy=False):
    """
    Read a scalar or array value from a process variable.

    :param chid:      Channel identifier
    :param chtype:    The external type of the supplied value to be written.
                      Conversion on the server will occur if this does not match the native type.
                      Default is the native type.
    :param count:     Element count to be read from the specified channel.
                      If *callback* is specified, a count of zero means use the current element count from the server.
    :param callback:  User supplied callback function to be run when requested operation completes.
                      The callback argument is a dict including the following fields:

                      ============   =============
                      field          value
                      ============   =============
                      chid           channel identifier
                      type           the type of the item returned, :class:`caffi.ca.DBR`
                      count          the element count of the item returned
                      status         status code of the request from the server, :class:`caffi.ca.ECA`
                      value          If *type* is a plain type, this is the PV's value. Otherwise it is a dict
                                     containing the meta information associated with this *type*.
                      ============   =============

    :param use_numpy: whether to format numeric waveform as numpy array
    :type chid:       cdata
    :type chtype:     int, :class:`caffi.ca.DBR`, None
    :type count:      int, None
    :type callback:   callable, None
    :type use_numpy:  bool
    :return:          (:class:`caffi.ca.ECA`, :class:`caffi.ca.DBRValue`)
    :rtype:           tuple

    When no *callback* is specified the returned channel value can't be assumed to be stable
    in the application supplied buffer until after :data:`caffi.ca.ECA.NORMAL` is returned from :func:`pend_io`.
    If a connection is lost outstanding ca get requests are not automatically reissued following reconnect.

    Call :meth:`caffi.ca.DBRValue.get` to retrieve the value.

    When *callback* is specified a value is read from the channel and
    then the user's callback is invoked with a dict containing the value.
    Note that ca_pend_io will not block for the delivery of values.
    If the channel disconnects before a ca get callback request can be completed,
    then the clients call back function is called with failure status.

    All of these functions return :data:`caffi.ca.ECA.DISCONN` if the channel is currently disconnected.

    All get requests are accumulated (buffered) and not forwarded to the IOC until one of
    :func:`flush_io`, :func:`pend_io`, or :func:`pend_event` are called.
    This allows several requests to be efficiently sent over the network in one message.

    """
    if chid not in __channels:
        return ECA.BADCHID, None

    if chtype is None:
        chtype = field_type(chid)
    if chtype == DBR.INVALID:
        return ECA.BADTYPE, None

    native_count = element_count(chid)
    if count is None or count <= 0 or count > native_count:
        count = native_count

    if callable(callback):
        get_callback = ffi.new_handle((callback, use_numpy))
        status = libca.ca_array_get_callback(chtype, count, chid, _get_callback, get_callback)
        if status == ECA.NORMAL:
            __channels[chid]['callbacks'].add(get_callback)
        return ECA(status), None
    else:
        value = ffi.new('char[]', dbr_size_n(chtype, count))
        status = libca.ca_array_get(chtype, count, chid, value)
        return ECA(status), DBRValue(chtype, count, value, use_numpy)


@ffi.callback('void(struct event_handler_args)')
def _put_callback(arg):
    epics_arg = {
        'chid':   arg.chid,
        'type':   DBR(arg.type),
        'count':  arg.count,
        'status': ECA(arg.status)
    }
    # If chid or the callback object is not in cache, it well indicates
    # that the python object has been garbage collected.
    # Then don't try to call from_handle, that is undefined and may crash.
    if arg.chid not in __channels or arg.usr not in __channels[arg.chid]['callbacks']:
        return

    user_callback = ffi.from_handle(arg.usr)
    __channels[arg.chid]['callbacks'].remove(arg.usr)
    if callable(user_callback):
        user_callback(epics_arg)


def _setup_put(chid, value, chtype=None, count=None):
    """
    Setup the C value for ca put. This is used by both :func:`put` and :func:`sg_put`.
    Return (chtype, count, c_value) tuple.
    """
    if chtype is None:
        chtype = field_type(chid)

    native_count = element_count(chid)
    if count is None or count <= 0 or count > native_count:
        count = native_count

    if isinstance(value, numbers.Number):
        value_count = 1
    elif isinstance(value, collections.Sequence):
        value_count = len(value)
        if isinstance(value, basestring):
            # convert to bytes
            value = to_bytes(value)
            if chtype == DBR.STRING:
                value_count = 1
            elif chtype == DBR.ENUM:
                chtype = DBR.STRING
                value_count = 1
            elif chtype == DBR.CHAR:
                value = [x for x in bytearray(value)] + [0]
                value_count = len(value)
            else:
                value = [float(value)]
                value_count = 1

        if value_count == 1 and chtype != DBR_STRING:
            value = value[0]
    else:
        return chtype, count, None

    # setup c value
    if value_count == 1:
        if chtype == DBR.STRING:
            value = to_bytes(value)
        cvalue = ffi.new(DBR_TYPE_STRING[chtype] + '*', value)
    else:
        if chtype == DBR.STRING:
            value = [to_bytes(x) for x in value]
        cvalue = ffi.new(DBR_TYPE_STRING[chtype] + '[]', value)

    # the actual count requested is the minimum of all three
    count = min(count, native_count, value_count)

    return chtype, count, cvalue


def put(chid, value, chtype=None, count=None, callback=None):
    """
    Write a scalar or array value to a process variable.

    :param chid:     Channel identifier
    :param value:    A scalar or array value to be written to the channel. If *value* is of string type, it will first
                     be convert to bytes using UTF8 codec. And the following conversion may be involved:

                     ============  =============
                     request type  conversion
                     ============  =============
                     DBR.STRING    nothing
                     DBR.ENUM      request type is changed to DBR.STRING
                     DBR.CHAR      a list of byte integers
                     Other types   a float number
                     ============  =============
    :param chtype:   The external type of the supplied value to be written.
                     Conversion on the server will occur if this does not match the native type.
                     Default is the native type.
    :param count:    Element count to be written to the channel. Default is native element count.
                     But it can be reduced to match the length of user supplied value.
    :param callback: User supplied callback function to be run when requested operation completes.
                     The callback argument is a dict including the following fields:

                     ============   =============
                     field          value
                     ============   =============
                     chid           channel identifier
                     type           DBR.INVALID (-1)
                     count          0
                     status         status code of the request from the server, :class:`caffi.ca.ECA`
                     ============   =============

    :type chid:      cdata
    :type value:     int, float, bytes, str, tuple, list, array
    :type chtype:    int, :class:`caffi.ca.DBR`, None
    :type count:     int, None
    :type callback:  callable, None
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.BADCHID - Corrupted CHID
        - ECA.BADTYPE - Invalid DBR_XXXX type
        - ECA.BADCOUNT - Requested count larger than native element count
        - ECA.STRTOBIG - Unusually large string supplied
        - ECA.NOWTACCESS - Write access denied
        - ECA.ALLOCMEM - Unable to allocate memory
        - ECA.DISCONN - Channel is disconnected

    When invoked without callback the client will receive no response unless the request
    can not be fulfilled in the server. If unsuccessful an exception handler is run on the client side.

    When invoked with callback the user supplied asynchronous call back is called only
    after the initiated write operation, and all actions resulting from the initiating write operation, complete.

    If unsuccessful the call back function is invoked indicating failure status.

    If the channel disconnects before a put callback request can be completed, then the client's call back function is
    called with failure status, but this does not guarantee that the server did not receive and process the request
    before the disconnect. If a connection is lost and then resumed outstanding ca put requests are not automatically
    reissued following reconnect.

    All of these functions return :data:`caffi.ca.ECA.DISCONN` if the channel is currently disconnected.

    All put requests are accumulated (buffered) and not forwarded to the IOC until
    one of :func:`flush_io`, :func:`pend_io`, or :func:`pend_event` are called.
    This allows several requests to be efficiently combined into one message.

    """
    if chid not in __channels:
        return ECA.BADCHID

    chtype, count, cvalue = _setup_put(chid, value, chtype, count)

    if cvalue is None:
        return ECA.BADTYPE

    if callback is None or not callable(callback):
        status = libca.ca_array_put(chtype, count, chid, cvalue)
    else:
        put_callback = ffi.new_handle(callback)
        __channels[chid]['callbacks'].add(put_callback)
        status = libca.ca_array_put_callback(chtype, count, chid, cvalue, _put_callback, put_callback)

    return ECA(status)


@ffi.callback('void(struct event_handler_args)')
def _event_callback(arg):
    # If chid or the callback object is not in cache, it well indicates
    # that the python object has been garbage collected.
    # Then don't try to call from_handle, that is undefined and may crash.
    if arg.chid not in __channels or arg.usr not in __channels[arg.chid]['monitors'].values():
        return

    user_callback, use_numpy = ffi.from_handle(arg.usr)

    epics_arg = {
        'chid':   arg.chid,
        'type':   DBR(arg.type),
        'count':  arg.count,
        'status': ECA(arg.status),
        'value':  format_dbr(arg.type, arg.count, arg.dbr, use_numpy)
    }
    if callable(user_callback):
        user_callback(epics_arg)


def create_subscription(chid, callback, chtype=None, count=None, mask=None, use_numpy=False):
    """
    Register a state change subscription and specify a call back function to be invoked
    whenever the process variable undergoes significant state changes.

    :param chid:      Channel identifier
    :param callback:  User supplied callback function to be run when requested operation completes.
                      The callback argument is a dict including the following fields:

                      ============   =============
                      field          value
                      ============   =============
                      chid           channel identifier
                      type           the type of the item returned, :class:`caffi.ca.DBR`
                      count          the element count of the item returned
                      status         status code of the request from the server, :class:`caffi.ca.ECA`
                      value          If *type* is a plain type, this is the PV's value. Otherwise it is a dict
                                     containing the meta information associated with this *type*.
                      ============   =============

    :param chtype:    The external type of the supplied value to be written.
                      Conversion will occur if this does not match the native type.
                      Default is the native type.
    :param count:     Element count to be written to the channel. Default is native element count.
    :param mask:      A mask with bits set for each of the event trigger types requested.
                      The event trigger mask must be a bitwise or of one or more of :class:`caffi.ca.DBE`.
    :param use_numpy: whether to format numeric waveform as numpy array
    :type chid:       cdata
    :type callback:   callable
    :type chtype:     :class:`caffi.ca.DBR`, None
    :type count:      int, None
    :type mask:       :class:`caffi.ca.DBE`, None
    :type use_numpy:  bool

    :return: (status code, event identifier)
    :rtype: (:class:`caffi.ca.ECA`, cdata)

    A significant change can be a change in the process variable's value, alarm status, or alarm severity.
    In the process control function block database the deadband field determines
    the magnitude of a significant change for the process variable's value.
    Each call to this function consumes resources in the client library and potentially a CA server
    until one of :func:`clear_channel` or :func:`clear_subscription` is called.

    Subscriptions may be installed or canceled against both connected and disconnected channels.
    The specified *callback* is called once immediately after the subscription is installed
    with the process variable's current state if the process variable is connected.
    Otherwise, the specified *callback* is called immediately after establishing a connection
    (or reconnection) with the process variable. The specified *callback* is called immediately
    with the process variable's current state from within :func:`create_subscription`
    if the client and the process variable share the same address space.

    If a subscription is installed on a channel in a disconnected state
    then the requested count will be set to the native maximum element count of the channel
    if the requested count is larger.

    All subscription requests such as the above are accumulated (buffered)
    and not forwarded to the IOC until one of :func:`flush_io`, :func:`pend_io`, or :func:`pend_event` are called.
    This allows several requests to be efficiently sent over the network in one message.

    If at any time after subscribing, read access to the specified process variable is lost,
    then the call back will be invoked immediately indicating that read access was lost via the status argument.
    When read access is restored normal event processing will resume starting always
    with at least one update indicating the current state of the channel.

    """
    if chid not in __channels:
        return ECA.BADCHID, None

    if chtype is None:
        chtype = field_type(chid)
    if chtype == DBR.INVALID:
        return ECA.BADTYPE, None

    # count = 0 is valid for subscription. It means only the number of changes elements.
    native_count = element_count(chid)
    if count is None or count < 0 or count > native_count:
        count = native_count

    if mask is None:
        mask = DBE.VALUE | DBE.ALARM

    pevid = ffi.new('evid *')

    monitor_callback = ffi.new_handle((callback, use_numpy))

    # - ECA.NORMAL - Normal successful completion
    # - ECA.BADCHID - Corrupted CHID
    # - ECA.BADTYPE - Invalid DBR_XXXX type
    # - ECA.ALLOCMEM - Unable to allocate memory
    # - ECA.ADDFAIL - A local database event add failed
    status = libca.ca_create_subscription(chtype, count, chid, mask, _event_callback, monitor_callback, pevid)
    if status != ECA_NORMAL:
        return ECA(status), None

    evid = pevid[0]
    __channels[chid]['monitors'][evid] = monitor_callback

    return ECA(status), evid


def clear_subscription(evid):
    """
    Cancel a subscription.

    :param cdata evid: event id returned by :meth:`create_subscription`
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.BADCHID - Corrupted CHID

    All cancel-subscription requests such as the above are accumulated (buffered) and not forwarded to the server
    until one of :func:`flush_io`, :func:`pend_io`, or :func:`pend_event` are called.
    This allows several requests to be efficiently sent together in one message.

    """
    status = libca.ca_clear_subscription(evid)
    chid = libca.ca_evid_to_chid(evid)

    if chid not in __channels:
        return ECA.BADCHID, None

    if evid in __channels[chid]['monitors']:
        del __channels[chid]['monitors'][evid]

    return ECA(status)


def clear_channel(chid):
    """
    Shutdown and reclaim resources associated with a channel created by ca_create_channel().

    :param cdata chid: Channel identifier
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.BADCHID - Corrupted CHID

    All remote operation requests such as the above are accumulated (buffered) and not forwarded to the IOC
    until one of :func:`flush_io`, :func:`pend_io` or :func:`pend_event` are called.
    This allows several requests to be efficiently sent over the network in one message.

    Clearing a channel does not cause its disconnect handler to be called,
    but clearing a channel does shutdown and reclaim any channel state change event subscriptions (monitors)
    registered with the channel.

    """
    if chid not in __channels:
        return ECA.BADCHID, None

    # clear all subscriptions for this channel
    for evid in list(__channels[chid]['monitors']):
        clear_subscription(evid)

    status = libca.ca_clear_channel(chid)

    # remove from channels list
    del __channels[chid]

    return ECA(status)


def pend_event(timeout):
    """
    The send buffer is flushed and CA background activity is processed for *timeout* seconds.

    :param float timeout: The duration to block in this routine in seconds. A timeout of zero seconds blocks forever.
    :return:
        - ECA.TIMEOUT - The operation timed out
        - ECA.EVDISALLOW - Function inappropriate for use within a call back handler


    The :func:`pend_event` function will not return before the specified timeout expires and
    all unfinished channel access labor has been processed, and unlike :func:`pend_io`
    returning from the function does not indicate anything about the status of pending IO requests.

    It return :data:`caffi.ca.ECA.TIMEOUT` when successful. This behavior probably isn't intuitive,
    but it is preserved to insure backwards compatibility.

    See also Thread Safety and Preemptive Callback to User Code.
    """
    status = libca.ca_pend_event(timeout)
    return ECA(status)


def poll():
    """
    The send buffer is flushed and any outstanding CA background activity is processed.

    .. note:: same as pend_event(1e-12)

    """
    status = libca.ca_pend_event(1e-12)
    return ECA(status)


def pend_io(timeout):
    """
    This function flushes the send buffer and then blocks until outstanding get requests *without* callback complete,
    and until channels created *without* callback connect for the first time.

    :param float timeout: Specifies the time out interval. A timeout interval of zero specifies forever.
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.TIMEOUT - Selected IO requests didn't complete before specified timeout
        - ECA.EVDISALLOW - Function inappropriate for use within an event handler

    If :data:`caffi.ca.ECA.NORMAL` is returned then it can be safely assumed that
    all outstanding get requests *without* callback have completed successfully and
    channels created *without* callback have connected for the first time.

    If :data:`caffi.ca.ECA.TIMEOUT` is returned then it must be assumed for all previous get requests and
    properly qualified first time channel connects have failed.

    If :data:`caffi.ca.ECA.TIMEOUT` is returned then get requests may be reissued
    followed by a subsequent call to :func:`pend_io`.
    Specifically, the function will block only for outstanding get requests issued,
    and also any channels created *without* callback,
    after the last call to ca_pend_io() or ca client context creation whichever is later.
    Note that :func:`create_channel` requests generally should not be reissued for the same process variable
    unless :func:`clear_channel` is called first.

    If no get or connection state change events are outstanding then :func:`pend_io` will flush the send buffer and
    return immediately without processing any outstanding channel access background activities.

    The delay specified should take into account worst case network delays
    such as Ethernet collision exponential back off until retransmission delays
    which can be quite long on overloaded networks.

    Unlike :func:`pend_event`, this routine will not process CA's background activities
    if none of the selected IO requests are pending.

    """
    status = libca.ca_pend_io(timeout)
    return ECA(status)


def pend(timeout, early):
    """
    :param float timeout: Specifies the time out interval. A timeout interval of zero specifies forever.
    :param bool early: Call :func:`pend_io` if *early* is True otherwise :func:`pend_event` is called
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.TIMEOUT - Selected IO requests didn't complete before specified timeout
        - ECA.EVDISALLOW - Function inappropriate for use within an event handler

    """
    status = libca.ca_pend(timeout, early)
    return ECA(status)


def test_io():
    """
    This function tests to see if all get requests are complete and channels created without
    a connection callback function are connected. It will report the status of outstanding get requests issued,
    and channels created without connection callback function,
    after the last call to ca_pend_io() or CA context initialization whichever is later.

    :return:
        - ECA.IODONE - All IO operations completed
        - ECA.IOINPROGRESS - IO operations still in progress
    """
    status = libca.ca_test_io()
    return ECA(status)


def flush_io():
    """
    Flush outstanding IO requests to the server.

    :return:
        - ECA.NORMAL - Normal successful completion

    This routine might be useful to users who need to flush requests prior to performing client side labor
    in parallel with labor performed in the server.
    Outstanding requests are also sent whenever the buffer which holds them becomes full.
    """
    status = libca.ca_flush_io()
    return ECA(status)


def field_type(chid):
    """
    :param cdata chid: channel identifier
    :return: the native type in the server of the process variable.
    :rtype: :class:`caffi.ca.DBF`
    """
    return DBF(libca.ca_field_type(chid))


def element_count(chid):
    """
    :param cdata chid: channel identifier
    :return: the maximum array element count in the server for the specified IO channel.
    """
    # ca_element_count returns long in C, and cffi convert to a `long` in Python.
    # In Python 2, the `long` type has a *L* suffix when displayed. This is disturbing since mostly this number will
    # not exceed the integer range. In Python 3, there is only a `long` type for integers and without *L* suffix.
    # Only for the sake of visual pleasure, the number is converted to `int` in Python 2. In any case, if the number
    # indeed exceeds 2*31-1 (on 32bit) or 2**63-1 (on 64bit), it gets up casted to `long`.
    count = libca.ca_element_count(chid)
    if sys.hexversion < 0x03000000:
        return int(count)
    else:
        return count


def name(chid):
    """
    :param cdata chid: channel identifier
    :return: the name provided when the supplied channel id was created.
    """
    return to_string(ffi.string(libca.ca_name(chid)))


def state(chid):
    """
    :param cdata chid: channel identifier
    :return: the connection state
    :rtype: :class:`caffi.ca.ChannelState`
    """
    if chid is None:
        return ChannelState.NEVER_SEARCH
    else:
        return ChannelState(libca.ca_state(chid))


def message(status):
    """
    :param int status: CA status code
    :return: a message corresponding to a user specified CA status code.
    """
    return to_string(ffi.string(libca.ca_message(status)))


def host_name(chid):
    """
    :param chid: channel identifier
    :return: the name of the host to which a channel is currently connected.
    """
    cstring = ffi.new('char[256]')
    libca.ca_get_host_name(chid, cstring, 256)
    return to_string(ffi.string(cstring))


def read_access(chid):
    """
    :param chid: channel identifier
    :return: True if the client currently has read access to the specified channel and False otherwise.
    """
    return libca.ca_read_access(chid) == 1


def write_access(chid):
    """
    :param chid: channel identifier
    :return: True if the client currently has write access to the specified channel and False otherwise.
    """
    return libca.ca_write_access(chid) == 1


def sg_create():
    """
    Create a synchronous group and return an identifier for it.

    :return: (:class:`caffi.ca.ECA`, Synchronous group identifier)
    :rtype: tuple

    A synchronous group can be used to guarantee that a set of channel access requests have completed.
    Once a synchronous group has been created then channel access get and put requests may be issued
    within it using :func:`sg_get` and :func:`sg_put` respectively.
    The routines :func:`sg_block` and :func:`sg_test` can be used to block for and test for completion respectively.
    The routine :func:`sg_reset` is used to discard knowledge of old requests which have timed out and
    in all likelihood will never be satisfied.

    Any number of asynchronous groups can have application requested operations outstanding
    within them at any given time.

    """
    pgid = ffi.new('CA_SYNC_GID *')
    status = libca.ca_sg_create(pgid)
    if status != ECA_NORMAL:
        return ECA(status), None
    else:
        gid = pgid[0]
        return ECA(status), gid


def sg_delete(gid):
    """
    Deletes a synchronous group.

    :param int gid: Identifier of the synchronous group to be deleted.
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.BADSYNCGRP - Invalid synchronous group
    """
    status = libca.ca_sg_delete(gid)
    return ECA(status)


def sg_block(gid, timeout):
    """
    Flushes the send buffer and then waits until outstanding requests complete or the specified time out expires.
    At this time outstanding requests include calls to :func:`sg_get` and calls to :func:`sg_put`.
    If :data:`caffi.ca.ECA.TIMEOUT` is returned then failure must be assumed for all outstanding queries.
    Operations can be reissued followed by another :func:`sg_block`.
    This routine will only block on outstanding queries issued after the last call to :func:`sg_block`,
    :func:`sg_reset`, or :func:`sg_create` whichever occurs later in time.
    If no queries are outstanding then :func:`sg_block` will return immediately
    without processing any pending channel access activities.

    Values written into your program's variables by a channel access synchronous group request should not be
    referenced by your program until :data:`caffi.ca.ECA.NORMAL` has been received from :func:`sg_block`.
    This routine will process pending channel access background activity while it is waiting.

    :param int gid:       Identifier of the synchronous group.
    :param float timeout: Specifies the time out interval. A timeout interval of zero specifies forever.
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.TIMEOUT - The operation timed out
        - ECA.EVDISALLOW - Function inappropriate for use within an event handler
        - ECA.BADSYNCGRP - Invalid synchronous group
    """
    status = libca.ca_sg_block(gid, timeout)
    return ECA(status)


def sg_test(gid):
    """
    Test to see if all requests made within a synchronous group have completed.

    :param int gid: Identifier of the synchronous group.
    :return:
        - ECA.IODONE - IO operations completed
        - ECA.IOINPROGRESS - Some IO operations still in progress
    """
    status = libca.ca_sg_test(gid)
    return ECA(status)


def sg_reset(gid):
    """
    Reset the number of outstanding requests within the specified synchronous group to zero so that
    :func:`sg_test` will return True and :func:`sg_block` will not block unless additional subsequent requests are made.

    :param int gid: Identifier of the synchronous group.
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.BADSYNCGRP - Invalid synchronous group
    """
    status = libca.sg_reset(gid)
    return ECA(status)


def sg_put(gid, chid, value, chtype=None, count=None):
    """
    Write a value, or array of values, to a channel and increment the outstanding request count of a synchronous group.

    :param gid:    Synchronous group identifier
    :param chid:   Channel identifier
    :param value:  The value or array of values to write. If *value* is of string type, it will first
                     be convert to bytes using UTF8 codec. And the following conversion may be involved:

                     ============  =============
                     request type  conversion
                     ============  =============
                     DBR.STRING    nothing
                     DBR.ENUM      request type is changed to DBR.STRING
                     DBR.CHAR      a list of byte integers
                     Other types   a float number
                     ============  =============

    :param chtype: The type of supplied value.
                   Conversion on the server will occur if it does not match the native type.
    :param count:  The element count to be written to the specified channel.
    :type gid:     int
    :type chid:    cdata
    :type value:   int, float, bytes, str, tuple, list, array
    :type chtype:  int, :class:`caffi.ca.DBR`
    :return:
        - ECA.NORMAL - Normal successful completion
        - ECA.BADSYNCGRP - Invalid synchronous group
        - ECA.BADCHID - Corrupted CHID
        - ECA.BADTYPE - Invalid DBR_XXXX type
        - ECA.BADCOUNT - Requested count larger than native element count
        - ECA.STRTOBIG - Unusually large string supplied
        - ECA.PUTFAIL - A local database put failed

    All remote operation requests such as the above are accumulated (buffered) and not forwarded to the server
    until one of :func:`flush_io`, :func:`pend_io` or :func:`pend_event` are called.
    This allows several requests to be efficiently sent in one message.

    If a connection is lost and then resumed outstanding puts are not reissued.
    """
    chtype, count, cvalue = _setup_put(chid, value, chtype, count)

    if cvalue is None:
        return ECA.BADTYPE

    status = libca.ca_sg_array_put(gid, chtype, count, chid, cvalue)

    return ECA(status)


def sg_get(gid, chid, chtype=None, count=None, use_numpy=False):
    """
    Read a value from a channel and increment the outstanding request count of a synchronous group.

    :param gid:       Identifier of the synchronous group.
    :param chid:      Channel identifier
    :param chtype:    External type of returned value.
                      Conversion on the server will occur if this does not match native type.
    :param count:     Element count to be read from the specified channel.
    :param use_numpy: whether to format numeric waveform as numpy array
    :type gid:        int
    :type chid:       cdata
    :type chtype:     int, :class:`caffi.ca.DBR`, None
    :type count:      int, None
    :return: (:class:`caffi.ca.ECA`, :class:`caffi.ca.DBRValue`)
    :rtype: tuple

    The values returned by :func:`sg_get` should not be referenced by your program
    until :data:`caffi.ca.ECA.NORMAL` has been received from ca_sg_block , or until :func:`sg_test` returns True.

    Call :meth:`caffi.ca.DBRValue.get` to retrieve the value.

    All remote operation requests such as the above are accumulated (buffered) and not forwarded to the server
    until one of :func:`flush_io`, :func:`pend_io`, or :func:`pend_event` are called.
    This allows several requests to be efficiently sent in one message.

    If a connection is lost and then resumed outstanding gets are not reissued.
    """
    native_count = element_count(chid)
    if count is None or count <= 0 or count > native_count:
        count = native_count

    if chtype is None:
        chtype = libca.ca_field_type(chid)

    cvalue = ffi.new('char[]', dbr_size_n(chtype, count))
    status = libca.ca_sg_array_get(gid, chtype, count, chid, cvalue)
    if status != ECA_NORMAL:
        return ECA(status), None
    else:
        return ECA(status), DBRValue(chtype, count, cvalue, use_numpy)


def version():
    """
    :return: CA version string
    :rtype: str
    """
    return to_string(ffi.string(libca.ca_version()))



import os
import platform
import sys
from cffi import FFI

__all__ = ['ffi', 'libca']

ffi = FFI()

# cadef.h
ffi.cdef("""
typedef void *chid;
typedef chid chanId;
typedef long chtype;
typedef double ca_real;
typedef void *evid;

/* arguments passed to user connection handlers */
struct connection_handler_args {
    chanId chid; /* channel id */
    long op;    /* one of CA_OP_CONN_UP or CA_OP_CONN_DOWN */
};
typedef void caCh (struct connection_handler_args args);

/* CFFI does not support bit field */
/*
typedef struct ca_access_rights {
    unsigned    read_access:1;
    unsigned    write_access:1;
} caar;
*/
typedef struct ca_access_rights {
    unsigned    access;
} caar;

/* arguments passed to user access rights handlers */
struct  access_rights_handler_args {
    chanId  chid;   /* channel id */
    caar    ar;     /* new access rights state */
};

typedef void caArh (struct access_rights_handler_args args);


/*
 * Arguments passed to event handlers and get/put call back handlers.
 *
 * The status field below is the CA ECA_XXX status of the requested
 * operation which is saved from when the operation was attempted in the
 * server and copied back to the clients call back routine.
 * If the status is not ECA_NORMAL then the dbr pointer will be NULL
 * and the requested operation can not be assumed to be successful.
 */
typedef struct event_handler_args {
    void            *usr;   /* user argument supplied with request */
    chanId          chid;   /* channel id */
    long            type;   /* the type of the item returned */
    long            count;  /* the element count of the item returned */
    const void      *dbr;   /* a pointer to the item returned */
    int             status; /* ECA_XXX status of the requested op from the server */
} evargs;
typedef void caEventCallBackFunc (struct event_handler_args);

void ca_test_event
(
    struct event_handler_args
);

/* arguments passed to user exception handlers */
struct exception_handler_args {
    void            *usr;   /* user argument supplied when installed */
    chanId          chid;   /* channel id (may be nill) */
    long            type;   /* type requested */
    long            count;  /* count requested */
    void            *addr;  /* user's address to write results of CA_OP_GET */
    long            stat;   /* channel access ECA_XXXX status code */
    long            op;     /* CA_OP_GET, CA_OP_PUT, ..., CA_OP_OTHER */
    const char      *ctx;   /* a character string containing context info */
    const char      *pFile; /* source file name (may be NULL) */
    unsigned        lineNo; /* source file line number (may be zero) */
};

typedef unsigned CA_SYNC_GID;

/*
 *  External OP codes for CA operations
 */
#define CA_OP_GET             0
#define CA_OP_PUT             1
#define CA_OP_CREATE_CHANNEL  2
#define CA_OP_ADD_EVENT       3
#define CA_OP_CLEAR_EVENT     4
#define CA_OP_OTHER           5

/*
 * used with connection_handler_args
 */
#define CA_OP_CONN_UP       6
#define CA_OP_CONN_DOWN     7

/* depricated */
#define CA_OP_SEARCH        2

short           ca_field_type(chid chan);
unsigned long   ca_element_count(chid chan);
const char *    ca_name (chid chan);
void            ca_set_puser (chid chan, void *puser);
void *          ca_puser (chid chan);
unsigned        ca_read_access (chid chan);
unsigned        ca_write_access (chid chan);

/*
 *  cs_ - `channel state'
 *
 *  cs_never_conn       valid chid, IOC not found
 *  cs_prev_conn        valid chid, IOC was found, but unavailable
 *  cs_conn             valid chid, IOC was found, still available
 *  cs_closed           channel deleted by user
 */
enum channel_state {cs_never_conn, cs_prev_conn, cs_conn, cs_closed};
enum channel_state ca_state (chid chan);

/************************************************************************/
/*  Perform Library Initialization                                      */
/*                                                                      */
/*  Must be called once before calling any of the other routines        */
/************************************************************************/
enum ca_preemptive_callback_select
{ ca_disable_preemptive_callback, ca_enable_preemptive_callback };
int ca_context_create(enum ca_preemptive_callback_select enable_premptive);
void ca_detach_context ();

/************************************************************************/
/*  Remove CA facility from your task                                   */
/*                                                                      */
/*  Normally called automatically at task exit                          */
/************************************************************************/
void ca_context_destroy (void);

typedef unsigned capri;
#define CA_PRIORITY_MAX 99
#define CA_PRIORITY_MIN 0
#define CA_PRIORITY_DEFAULT 0

#define CA_PRIORITY_DB_LINKS 80
#define CA_PRIORITY_ARCHIVE 20
#define CA_PRIORITY_OPI 0

/*
 * ca_create_channel ()
 *
 * pChanName            R   channel name string
 * pConnStateCallback   R   address of connection state change
 *                          callback function
 * pUserPrivate         R   placed in the channel's user private field
 *                          o can be fetched later by ca_puser(CHID)
 *                          o passed as void * arg to *pConnectCallback above
 * priority             R   priority level in the server 0 - 100
 * pChanID              RW  channel id written here
 */
int ca_create_channel
(
    const char     *pChanName,
    caCh           *pConnStateCallback,
    void           *pUserPrivate,
    capri          priority,
    void           *pChanID
);

/*
 * ca_change_connection_event()
 *
 * chan     R   channel identifier
 * pfunc    R   address of connection call-back function
 */
int ca_change_connection_event
(
     chid       chan,
     caCh *     pfunc
);

/*
 * ca_replace_access_rights_event ()
 *
 * chan     R   channel identifier
 * pfunc    R   address of access rights call-back function
 */
int ca_replace_access_rights_event (
     chid   chan,
     caArh  *pfunc
);

/*
 * ca_add_exception_event ()
 *
 * replace the default exception handler
 *
 * pfunc    R   address of exception call-back function
 * pArg     R   copy of this pointer passed to exception
 *          call-back function
 */
typedef void caExceptionHandler (struct exception_handler_args);
int ca_add_exception_event
(
     caExceptionHandler *pfunc,
     void               *pArg
);

/*
 * ca_clear_channel()
 * - deallocate resources reserved for a channel
 *
 * chanId   R   channel ID
 */
int ca_clear_channel
(
     chid   chanId
);

/************************************************************************/
/*  Write a value to a channel                                          */
/************************************************************************/
/*
 * ca_array_put()
 *
 * type         R   data type from db_access.h
 * count        R   array element count
 * chan         R   channel identifier
 * pValue       R   new channel value copied from this location
 */
int ca_array_put
(
     chtype         type,
     unsigned long  count,
     chid           chanId,
     const void *   pValue
);

/*
 * ca_array_put_callback()
 *
 * This routine functions identically to the original ca put request
 * with the addition of a callback to the user supplied function
 * after recod processing completes in the IOC. The arguments
 * to the user supplied callback function are declared in
 * the structure event_handler_args and include the pointer
 * sized user argument supplied when ca_array_put_callback() is called.
 *
 * type         R   data type from db_access.h
 * count        R   array element count
 * chan         R   channel identifier
 * pValue       R   new channel value copied from this location
 * pFunc        R   pointer to call-back function
 * pArg         R   copy of this pointer passed to pFunc
 */
int ca_array_put_callback
(
     chtype                 type,
     unsigned long          count,
     chid                   chanId,
     const void *           pValue,
     caEventCallBackFunc *  pFunc,
     void *                 pArg
);

/************************************************************************/
/*  Read a value from a channel                                         */
/************************************************************************/
/*
 * ca_array_get()
 *
 * type     R   data type from db_access.h
 * count    R   array element count
 * chan     R   channel identifier
 * pValue   W   channel value copied to this location
 */
int ca_array_get
(
    long           type,
    unsigned long  count,
    chid           chanId,
    void *         pValue
);

/************************************************************************/
/*  Read a value from a channel and run a callback when the value       */
/*  returns                                                             */
/*                                                                      */
/*                                                                      */
/************************************************************************/
/*
 * ca_array_get_callback()
 *
 * type     R   data type from db_access.h
 * count    R   array element count
 * chan     R   channel identifier
 * pFunc    R   pointer to call-back function
 * pArg     R   copy of this pointer passed to pFunc
 */
int ca_array_get_callback
(
    chtype                 type,
    unsigned long          count,
    chid                   chanId,
    caEventCallBackFunc *  pFunc,
    void *                 pArg
);

/************************************************************************/
/*  Specify a function to be executed whenever significant changes      */
/*  occur to a channel.                                                 */
/*  NOTES:                                                              */
/*  1)  Evid may be omited by passing a NULL pointer                    */
/*                                                                      */
/*  2)  An array count of zero specifies the native db count            */
/*                                                                      */
/************************************************************************/

/*
 * ca_create_subscription ()
 *
 * type     R   data type from db_access.h
 * count    R   array element count
 * chan     R   channel identifier
 * mask     R   event mask - one of {DBE_VALUE, DBE_ALARM, DBE_LOG}
 * pFunc    R   pointer to call-back function
 * pArg     R   copy of this pointer passed to pFunc
 * pEventID W   event id written at specified address
 */
int ca_create_subscription
(
    chtype                 type,
    unsigned long          count,
    chid                   chanId,
    long                   mask,
    caEventCallBackFunc *  pFunc,
    void *                 pArg,
    evid *                 pEventID
);

/************************************************************************/
/*  Remove a function from a list of those specified to run             */
/*  whenever significant changes occur to a channel                     */
/*                                                                      */
/************************************************************************/
/*
 * ca_clear_subscription()
 *
 * eventID  R   event id
 */
int ca_clear_subscription
(
    evid    eventId
);

chid ca_evid_to_chid ( evid id );


/************************************************************************/
/*                                                                      */
/*   Requested data is not necessarily stable prior to                  */
/*   return from called subroutine.  Call ca_pend_io()                  */
/*   to guarantee that requested data is stable.  Call the routine      */
/*   ca_flush_io() to force all outstanding requests to be              */
/*   sent out over the network.  Significant increases in               */
/*   performance have been measured when batching several remote        */
/*   requests together into one message.  Additional                    */
/*   improvements can be obtained by performing local processing        */
/*   in parallel with outstanding remote processing.                    */
/*                                                                      */
/*   FLOW OF TYPICAL APPLICATION                                        */
/*                                                                      */
/*   search()       ! Obtain Channel ids                                */
/*   .              ! "				                                    */
/*   .              ! "                                                 */
/*   pend_io        ! wait for channels to connect                      */
/*                                                                      */
/*   get()          ! several requests for remote info                  */
/*   get()          ! "					                                */
/*   add_event()    ! "					                                */
/*   get()          ! "					                                */
/*   .                                                                  */
/*   .                                                                  */
/*   .                                                                  */
/*   flush_io()     ! send get requests                                 */
/*                  ! optional parallel processing                      */
/*   .              ! "					                                */
/*   .              ! "					                                */
/*   pend_io()      ! wait for replies from get requests                */
/*   .              ! access to requested data                          */
/*   .              ! "					                                */
/*   pend_event()   ! wait for requested events                         */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*  These routines wait for channel subscription events and call the    */
/*  functions specified with add_event when events occur. If the        */
/*  timeout is specified as 0 an infinite timeout is assumed.           */
/*  ca_flush_io() is called by this routine. If ca_pend_io ()           */
/*  is called when no IO is outstanding then it will return immediately */
/*  without processing.                                                 */
/************************************************************************/

/*
 * ca_pend_event()
 *
 * timeOut  R   wait for this delay in seconds
 */
int ca_pend_event(ca_real timeOut);

/*
 * ca_pend_io()
 *
 * timeOut  R   wait for this delay in seconds but return early
 *              if all get requests (or search requests with null
 *              connection handler pointer have completed)
 */
int ca_pend_io(ca_real timeOut);

/* calls ca_pend_io() if early is true otherwise ca_pend_event() is called */
int ca_pend (ca_real timeout, int early);

/*
 * ca_test_io()
 *
 * returns TRUE when get requests (or search requests with null
 * connection handler pointer) are outstanding
 */
int ca_test_io (void);

/************************************************************************/
/*  Send out all outstanding messages in the send queue                 */
/************************************************************************/
/*
 * ca_flush_io()
 */
int ca_flush_io();

/*
 * ca_host_name_function()
 *
 * channel  R   channel identifier
 *
 * !!!! this function is _not_ thread safe !!!!
 */
const char * ca_host_name (chid channel);
/* thread safe version */
unsigned ca_get_host_name ( chid pChan,
    char *pBuf, unsigned bufLength );

/*
 * ca_replace_printf_handler ()
 *
 * for apps that want to change where ca formatted
 * text output goes
 *
 * use two ifdef's for trad C compatibility
 *
 * ca_printf_func   R   pointer to new function called when
 *                          CA prints an error message
 */
/*
typedef int caPrintfFunc (const char *pformat, va_list args);
int ca_replace_printf_handler (
    caPrintfFunc    *ca_printf_func
);
*/

/*
 * CA synch groups
 *
 * This facility will allow the programmer to create
 * any number of synchronization groups. The programmer might then
 * interleave IO requests within any of the groups. Once The
 * IO operations are initiated then the programmer is free to
 * block for IO completion within any one of the groups as needed.
 */

/*
 * ca_sg_create()
 *
 * create a sync group
 *
 * pgid     W   pointer to sync group id that will be written
 */
int ca_sg_create (CA_SYNC_GID *  pgid);

/*
 * ca_sg_delete()
 *
 * delete a sync group
 *
 * gid      R   sync group id
 */
int ca_sg_delete (const CA_SYNC_GID gid);

/*
 * ca_sg_block()
 *
 * block for IO performed within a sync group to complete
 *
 * gid      R   sync group id
 * timeout  R   wait for this duration prior to timing out
 *          and returning ECA_TIMEOUT
 */
int ca_sg_block (const CA_SYNC_GID gid, ca_real timeout);

/*
 * ca_sg_test()
 *
 * test for sync group IO operations in progress
 *
 * gid      R   sync group id
 *
 * returns one of ECA_BADSYNCGRP, ECA_IOINPROGRESS, ECA_IODONE
 */
int ca_sg_test (const CA_SYNC_GID gid);

/*
 * ca_sg_reset
 *
 * gid      R   sync group id
 */
int ca_sg_reset(const CA_SYNC_GID gid);

/*
 * ca_sg_array_get()
 *
 * initiate a get within a sync group
 * (essentially a ca_array_get() with a sync group specified)
 *
 * gid      R   sync group id
 * type     R   data type from db_access.h
 * count    R   array element count
 * chan     R   channel identifier
 * pValue   W   channel value copied to this location
 */
int ca_sg_array_get
(
    const CA_SYNC_GID gid,
    chtype type,
    unsigned long count,
    chid chan,
    void *pValue
);

/*
 * ca_sg_array_put()
 *
 * initiate a put within a sync group
 * (essentially a ca_array_put() with a sync group specified)
 *
 * gid      R   sync group id
 * type     R   data type from db_access.h
 * count    R   array element count
 * chan     R   channel identifier
 * pValue   R   new channel value copied from this location
 */
int ca_sg_array_put
(
    const CA_SYNC_GID gid,
    chtype type,
    unsigned long count,
    chid chan,
    const void *pValue
);

/*
 * ca_sg_stat()
 *
 * print status of a sync group
 *
 * gid      R   sync group id
 */
int ca_sg_stat (CA_SYNC_GID gid);


/*
 * used when an auxillary thread needs to join a CA client context started
 * by another thread
 */
struct ca_client_context * ca_current_context ();
int ca_attach_context ( struct ca_client_context * context );


int ca_client_status ( unsigned level );
int ca_context_status ( struct ca_client_context *, unsigned level );


const char * ca_message(long ca_status);

/*
 * ca_version()
 *
 * returns the CA version string
 */
const char * ca_version (void);
""")

# alarm.h
ffi.cdef("""
#define NO_ALARM            0

/* ALARM SEVERITIES - must match menuAlarmSevr.dbd */
typedef enum {
    epicsSevNone = NO_ALARM,
    epicsSevMinor,
    epicsSevMajor,
    epicsSevInvalid,
    ALARM_NSEV
} epicsAlarmSeverity;

/* ALARM STATUS - must match menuAlarmStat.dbd */

typedef enum {
    epicsAlarmNone = NO_ALARM,
    epicsAlarmRead,
    epicsAlarmWrite,
    epicsAlarmHiHi,
    epicsAlarmHigh,
    epicsAlarmLoLo,
    epicsAlarmLow,
    epicsAlarmState,
    epicsAlarmCos,
    epicsAlarmComm,
    epicsAlarmTimeout,
    epicsAlarmHwLimit,
    epicsAlarmCalc,
    epicsAlarmScan,
    epicsAlarmLink,
    epicsAlarmSoft,
    epicsAlarmBadSub,
    epicsAlarmUDF,
    epicsAlarmDisable,
    epicsAlarmSimm,
    epicsAlarmReadAccess,
    epicsAlarmWriteAccess,
    ALARM_NSTATUS
} epicsAlarmCondition;
""")

# caeventmask.h
ffi.cdef("""
#define DBE_VALUE    1
#define DBE_ARCHIVE  2
#define DBE_LOG      2
#define DBE_ALARM    4
#define DBE_PROPERTY 8
""")

# epicsTypes.h
ffi.cdef("""
typedef int8_t          epicsInt8;
typedef uint8_t         epicsUInt8;
typedef int16_t         epicsInt16;
typedef uint16_t        epicsUInt16;
typedef epicsUInt16     epicsEnum16;
typedef int32_t         epicsInt32;
typedef uint32_t        epicsUInt32;
typedef int64_t         epicsInt64;
typedef uint64_t        epicsUInt64;

typedef float           epicsFloat32;
typedef double          epicsFloat64;
typedef epicsInt32      epicsStatus;

typedef struct {
    unsigned    length;
    char        *pString;
}epicsString;

/*
 * !! Dont use this - it may vanish in the future !!
 *
 * Provided only for backwards compatibility with
 * db_access.h
 *
 */
typedef char            epicsOldString[40];

""")

# epicsTime.h
ffi.cdef("""
/* epics time stamp for C interface*/
typedef struct epicsTimeStamp {
    epicsUInt32    secPastEpoch;   /* seconds since 0000 Jan 1, 1990 */
    epicsUInt32    nsec;           /* nanoseconds within second */
} epicsTimeStamp;
""")

# db_access.h
ffi.cdef("""
/*
 * architecture independent types
 *
 * (so far this is sufficient for all archs we have ported to)
 */
typedef epicsOldString dbr_string_t;
typedef epicsUInt8 dbr_char_t;
typedef epicsInt16 dbr_short_t;
typedef epicsUInt16	dbr_ushort_t;
typedef epicsInt16 dbr_int_t;
typedef epicsUInt16 dbr_enum_t;
typedef epicsInt32 dbr_long_t;
typedef epicsUInt32 dbr_ulong_t;
typedef epicsFloat32 dbr_float_t;
typedef epicsFloat64 dbr_double_t;
typedef epicsUInt16	dbr_put_ackt_t;
typedef epicsUInt16	dbr_put_acks_t;
typedef epicsOldString dbr_stsack_string_t;
typedef epicsOldString dbr_class_name_t;


/* VALUES WITH STATUS STRUCTURES */

/* structure for a  string status field */
struct dbr_sts_string {
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_string_t	value;			/* current value */
};

/* structure for a  string status and ack field */
struct dbr_stsack_string{
	dbr_ushort_t	status;	 		/* status of value */
	dbr_ushort_t	severity;		/* severity of alarm */
	dbr_ushort_t	ackt;	 		/* ack transient? */
	dbr_ushort_t	acks;			/* ack severity	*/
	dbr_string_t	value;			/* current value */
};
/* structure for an short status field */
struct dbr_sts_int{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	value;			/* current value */
};
struct dbr_sts_short{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	value;			/* current value */
};

/* structure for a  float status field */
struct dbr_sts_float{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_float_t	value;			/* current value */
};

/* structure for a  enum status field */
struct dbr_sts_enum{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_enum_t	value;			/* current value */
};

/* structure for a char status field */
struct dbr_sts_char{
	dbr_short_t	status;	 	/* status of value */
	dbr_short_t	severity;	/* severity of alarm */
	dbr_char_t	RISC_pad;	/* RISC alignment */
	dbr_char_t	value;		/* current value */
};

/* structure for a long status field */
struct dbr_sts_long{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_long_t	value;			/* current value */
};

/* structure for a double status field */
struct dbr_sts_double{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_long_t	RISC_pad;		/* RISC alignment */
	dbr_double_t	value;			/* current value */
};

/* VALUES WITH STATUS AND TIME STRUCTURES */

/* structure for a  string time field */
struct dbr_time_string{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_string_t	value;			/* current value */
};

/* structure for an short time field */
struct dbr_time_int{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_short_t	RISC_pad;		/* RISC alignment */
	dbr_short_t	value;			/* current value */
};

struct dbr_time_short{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_short_t	RISC_pad;		/* RISC alignment */
	dbr_short_t	value;			/* current value */
};

/* structure for a  float time field */
struct dbr_time_float{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_float_t	value;			/* current value */
};

/* structure for a  enum time field */
struct dbr_time_enum{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_short_t	RISC_pad;		/* RISC alignment */
	dbr_enum_t	value;			/* current value */
};

/* structure for a char time field */
struct dbr_time_char{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_short_t	RISC_pad0;		/* RISC alignment */
	dbr_char_t	RISC_pad1;		/* RISC alignment */
	dbr_char_t	value;			/* current value */
};

/* structure for a long time field */
struct dbr_time_long{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_long_t	value;			/* current value */
};

/* structure for a double time field */
struct dbr_time_double{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	epicsTimeStamp	stamp;			/* time stamp */
	dbr_long_t	RISC_pad;		/* RISC alignment */
	dbr_double_t	value;			/* current value */
};

/* VALUES WITH STATUS AND GRAPHIC STRUCTURES */

/* structure for a graphic string */
	/* not implemented; use struct_dbr_sts_string */

/* structure for a graphic short field */
struct dbr_gr_int{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_short_t	upper_disp_limit;	/* upper limit of graph */
	dbr_short_t	lower_disp_limit;	/* lower limit of graph */
	dbr_short_t	upper_alarm_limit;
	dbr_short_t	upper_warning_limit;
	dbr_short_t	lower_warning_limit;
	dbr_short_t	lower_alarm_limit;
	dbr_short_t	value;			/* current value */
};
struct dbr_gr_short{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_short_t	upper_disp_limit;	/* upper limit of graph */
	dbr_short_t	lower_disp_limit;	/* lower limit of graph */
	dbr_short_t	upper_alarm_limit;
	dbr_short_t	upper_warning_limit;
	dbr_short_t	lower_warning_limit;
	dbr_short_t	lower_alarm_limit;
	dbr_short_t	value;			/* current value */
};

/* structure for a graphic floating point field */
struct dbr_gr_float{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	precision;		/* number of decimal places */
	dbr_short_t	RISC_pad0;		/* RISC alignment */
	char		units[8];	/* units of value */
	dbr_float_t	upper_disp_limit;	/* upper limit of graph */
	dbr_float_t	lower_disp_limit;	/* lower limit of graph */
	dbr_float_t	upper_alarm_limit;
	dbr_float_t	upper_warning_limit;
	dbr_float_t	lower_warning_limit;
	dbr_float_t	lower_alarm_limit;
	dbr_float_t	value;			/* current value */
};

/* structure for a graphic enumeration field */
struct dbr_gr_enum{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	no_str;			/* number of strings */
	char		strs[16][26];
						/* state strings */
	dbr_enum_t	value;			/* current value */
};

/* structure for a graphic char field */
struct dbr_gr_char{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_char_t	upper_disp_limit;	/* upper limit of graph */
	dbr_char_t	lower_disp_limit;	/* lower limit of graph */
	dbr_char_t	upper_alarm_limit;
	dbr_char_t	upper_warning_limit;
	dbr_char_t	lower_warning_limit;
	dbr_char_t	lower_alarm_limit;
	dbr_char_t	RISC_pad;		/* RISC alignment */
	dbr_char_t	value;			/* current value */
};

/* structure for a graphic long field */
struct dbr_gr_long{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_long_t	upper_disp_limit;	/* upper limit of graph */
	dbr_long_t	lower_disp_limit;	/* lower limit of graph */
	dbr_long_t	upper_alarm_limit;
	dbr_long_t	upper_warning_limit;
	dbr_long_t	lower_warning_limit;
	dbr_long_t	lower_alarm_limit;
	dbr_long_t	value;			/* current value */
};

/* structure for a graphic double field */
struct dbr_gr_double{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	precision;		/* number of decimal places */
	dbr_short_t	RISC_pad0;		/* RISC alignment */
	char		units[8];	/* units of value */
	dbr_double_t	upper_disp_limit;	/* upper limit of graph */
	dbr_double_t	lower_disp_limit;	/* lower limit of graph */
	dbr_double_t	upper_alarm_limit;
	dbr_double_t	upper_warning_limit;
	dbr_double_t	lower_warning_limit;
	dbr_double_t	lower_alarm_limit;
	dbr_double_t	value;			/* current value */
};

/* VALUES WITH STATUS, GRAPHIC and CONTROL STRUCTURES */

/* structure for a control string */
	/* not implemented; use struct_dbr_sts_string */

/* structure for a control integer */
struct dbr_ctrl_int{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_short_t	upper_disp_limit;	/* upper limit of graph */
	dbr_short_t	lower_disp_limit;	/* lower limit of graph */
	dbr_short_t	upper_alarm_limit;
	dbr_short_t	upper_warning_limit;
	dbr_short_t	lower_warning_limit;
	dbr_short_t	lower_alarm_limit;
	dbr_short_t	upper_ctrl_limit;	/* upper control limit */
	dbr_short_t	lower_ctrl_limit;	/* lower control limit */
	dbr_short_t	value;			/* current value */
};
struct dbr_ctrl_short{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_short_t	upper_disp_limit;	/* upper limit of graph */
	dbr_short_t	lower_disp_limit;	/* lower limit of graph */
	dbr_short_t	upper_alarm_limit;
	dbr_short_t	upper_warning_limit;
	dbr_short_t	lower_warning_limit;
	dbr_short_t	lower_alarm_limit;
	dbr_short_t	upper_ctrl_limit;	/* upper control limit */
	dbr_short_t	lower_ctrl_limit;	/* lower control limit */
	dbr_short_t	value;			/* current value */
};

/* structure for a control floating point field */
struct dbr_ctrl_float{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	precision;		/* number of decimal places */
	dbr_short_t	RISC_pad;		/* RISC alignment */
	char		units[8];	/* units of value */
	dbr_float_t	upper_disp_limit;	/* upper limit of graph */
	dbr_float_t	lower_disp_limit;	/* lower limit of graph */
	dbr_float_t	upper_alarm_limit;
	dbr_float_t	upper_warning_limit;
	dbr_float_t	lower_warning_limit;
	dbr_float_t	lower_alarm_limit;
 	dbr_float_t	upper_ctrl_limit;	/* upper control limit */
	dbr_float_t	lower_ctrl_limit;	/* lower control limit */
	dbr_float_t	value;			/* current value */
};

/* structure for a control enumeration field */
struct dbr_ctrl_enum{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	no_str;			/* number of strings */
	char	strs[16][26];
					/* state strings */
	dbr_enum_t	value;		/* current value */
};

/* structure for a control char field */
struct dbr_ctrl_char{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_char_t	upper_disp_limit;	/* upper limit of graph */
	dbr_char_t	lower_disp_limit;	/* lower limit of graph */
	dbr_char_t	upper_alarm_limit;
	dbr_char_t	upper_warning_limit;
	dbr_char_t	lower_warning_limit;
	dbr_char_t	lower_alarm_limit;
	dbr_char_t	upper_ctrl_limit;	/* upper control limit */
	dbr_char_t	lower_ctrl_limit;	/* lower control limit */
	dbr_char_t	RISC_pad;		/* RISC alignment */
	dbr_char_t	value;			/* current value */
};

/* structure for a control long field */
struct dbr_ctrl_long{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	char		units[8];	/* units of value */
	dbr_long_t	upper_disp_limit;	/* upper limit of graph */
	dbr_long_t	lower_disp_limit;	/* lower limit of graph */
	dbr_long_t	upper_alarm_limit;
	dbr_long_t	upper_warning_limit;
	dbr_long_t	lower_warning_limit;
	dbr_long_t	lower_alarm_limit;
	dbr_long_t	upper_ctrl_limit;	/* upper control limit */
	dbr_long_t	lower_ctrl_limit;	/* lower control limit */
	dbr_long_t	value;			/* current value */
};

/* structure for a control double field */
typedef struct dbr_ctrl_double{
	dbr_short_t	status;	 		/* status of value */
	dbr_short_t	severity;		/* severity of alarm */
	dbr_short_t	precision;		/* number of decimal places */
	dbr_short_t	RISC_pad0;		/* RISC alignment */
	char		units[8];	/* units of value */
	dbr_double_t	upper_disp_limit;	/* upper limit of graph */
	dbr_double_t	lower_disp_limit;	/* lower limit of graph */
	dbr_double_t	upper_alarm_limit;
	dbr_double_t	upper_warning_limit;
	dbr_double_t	lower_warning_limit;
	dbr_double_t	lower_alarm_limit;
	dbr_double_t	upper_ctrl_limit;	/* upper control limit */
	dbr_double_t	lower_ctrl_limit;	/* lower control limit */
	dbr_double_t	value;			/* current value */
};

/* union for each fetch buffers */
union db_access_val{
	dbr_string_t		strval;		/* string max size	      */
	dbr_short_t		shrtval;	/* short		      */
	dbr_short_t		intval;		/* short		      */
	dbr_float_t		fltval;		/* IEEE Float		      */
	dbr_enum_t		enmval;		/* item number		      */
	dbr_char_t		charval;	/* character		      */
	dbr_long_t		longval;	/* long			      */
	dbr_double_t		doubleval;	/* double		      */
	struct dbr_sts_string	sstrval;	/* string field	with status   */
	struct dbr_sts_short	sshrtval;	/* short field with status    */
	struct dbr_sts_float	sfltval;	/* float field with status    */
	struct dbr_sts_enum	senmval;	/* item number with status    */
	struct dbr_sts_char	schrval;	/* char field with status     */
	struct dbr_sts_long	slngval;	/* long field with status     */
	struct dbr_sts_double	sdblval;	/* double field with time     */
	struct dbr_time_string	tstrval;	/* string field	with time     */
	struct dbr_time_short	tshrtval;	/* short field with time      */
	struct dbr_time_float	tfltval;	/* float field with time      */
	struct dbr_time_enum	tenmval;	/* item number with time      */
	struct dbr_time_char	tchrval;	/* char field with time	      */
	struct dbr_time_long	tlngval;	/* long field with time	      */
	struct dbr_time_double	tdblval;	/* double field with time     */
	struct dbr_sts_string	gstrval;	/* graphic string info	      */
	struct dbr_gr_short	gshrtval;	/* graphic short info	      */
	struct dbr_gr_float	gfltval;	/* graphic float info	      */
	struct dbr_gr_enum	genmval;	/* graphic item info	      */
	struct dbr_gr_char	gchrval;	/* graphic char info	      */
	struct dbr_gr_long	glngval;	/* graphic long info	      */
	struct dbr_gr_double	gdblval;	/* graphic double info	      */
	struct dbr_sts_string	cstrval;	/* control string info	      */
	struct dbr_ctrl_short	cshrtval;	/* control short info	      */
	struct dbr_ctrl_float	cfltval;	/* control float info	      */
	struct dbr_ctrl_enum	cenmval;	/* control item info	      */
	struct dbr_ctrl_char	cchrval;	/* control char info	      */
	struct dbr_ctrl_long	clngval;	/* control long info	      */
	struct dbr_ctrl_double	cdblval;	/* control double info	      */
	dbr_put_ackt_t		putackt;	/* item number		      */
	dbr_put_acks_t		putacks;	/* item number		      */
	struct dbr_sts_string	sastrval;	/* string field	with status   */
	dbr_string_t		classname;	/* string max size	      */
};

/* size for each type - array indexed by the DBR_ type code */
const unsigned short dbr_size[];

/* size for each type's value - array indexed by the DBR_ type code */
const unsigned short dbr_value_size[];

const unsigned short dbr_value_offset[];

const char   *dbf_text[];
const short  dbf_text_dim;
const char   *dbf_text_invalid;

const char   *dbr_text[];
const short  dbr_text_dim;
const char   *dbr_text_invalid;
""")


def get_libca():
    cwd = os.path.dirname(os.path.abspath(__file__))
    osname = platform.system()
    is64bit = sys.maxsize > 2**32
    flags = 0
    if osname == 'Darwin':
        host_arch = 'darwin-x86'
        libca_name = 'libca.dylib'
    elif osname == 'Linux':
        libca_name = 'libca.so'
        flags = ffi.RTLD_NODELETE
        if is64bit:
            host_arch = 'linux-x86_64'
        else:
            host_arch = 'linux-x86'
    elif osname == 'Windows':
        libca_name = 'ca.dll'
        if is64bit:
            host_arch = 'windows-x64'
        else:
            host_arch = 'win32-x86'
    else:
        raise OSError('Unsupported Operation System %s' % osname)

    # system defined epics installation precedes package bounded library
    # but not for Windows because the package bounded library has been
    # modified not to use __stdcall convention.
    epics_base = os.environ.get('EPICS_BASE')
    if epics_base and host_arch not in ['win32-x86', 'windows-x64']:
        host_arch = os.environ.get('EPICS_HOST_ARCH', host_arch)
        return os.path.join(epics_base, 'lib', host_arch), libca_name, flags
    else:
        return os.path.join(cwd, 'lib', host_arch), libca_name, flags

#
libca_path, libca_name, libca_flags = get_libca()

# save and set current dir to ca library's path
old_cwd = os.getcwd()
os.chdir(libca_path)

# load ca library
libca = ffi.dlopen(os.path.join(libca_path, libca_name), libca_flags)

# restore current dir
os.chdir(old_cwd)



from __future__ import print_function
import pytest
import threading
import caffi.ca as ca

def setup_module(module):
    global chid
    # create context
    status = ca.create_context(True)
    assert status == ca.ECA.NORMAL

    # create channel
    status, chid = ca.create_channel('catest')
    assert status == ca.ECA.NORMAL

    # wait for connection
    status = ca.pend_io(2)
    assert status == ca.ECA.NORMAL

    # write the value with callback
    # so that we are sure the new value has reached the server
    put_done = threading.Event()
    def put_callback(args):
        put_done.set()

    status = ca.put(chid, 1., callback=put_callback)
    assert status == ca.ECA.NORMAL

    ca.flush_io()
    put_done.wait()

@pytest.mark.parametrize("dbrtype", [
    ca.DBR.STRING, ca.DBR.SHORT, ca.DBR.FLOAT, ca.DBR.ENUM, ca.DBR.CHAR, ca.DBR.LONG, ca.DBR.DOUBLE,
    ca.DBR.STS_STRING, ca.DBR.STS_SHORT, ca.DBR.STS_FLOAT, ca.DBR.STS_ENUM,
    ca.DBR.STS_CHAR, ca.DBR.STS_LONG, ca.DBR.STS_DOUBLE,
    ca.DBR.TIME_STRING, ca.DBR.TIME_SHORT, ca.DBR.TIME_FLOAT, ca.DBR.TIME_ENUM,
    ca.DBR.TIME_CHAR, ca.DBR.TIME_LONG, ca.DBR.TIME_DOUBLE,
    ca.DBR.GR_STRING, ca.DBR.GR_SHORT, ca.DBR.GR_FLOAT, ca.DBR.GR_ENUM,
    ca.DBR.GR_CHAR, ca.DBR.GR_LONG, ca.DBR.GR_DOUBLE,
    ca.DBR.CTRL_STRING, ca.DBR.CTRL_SHORT, ca.DBR.CTRL_FLOAT, ca.DBR.CTRL_ENUM,
    ca.DBR.CTRL_CHAR, ca.DBR.CTRL_LONG, ca.DBR.CTRL_DOUBLE,
    ca.DBR.STSACK_STRING, ca.DBR.CLASS_NAME
    ])
def test_dbrtype(dbrtype):
    global chid
    status, dbrvalue = ca.get(chid, dbrtype)
    assert status == ca.ECA.NORMAL

    status = ca.pend_io(2)
    assert status == ca.ECA.NORMAL

    value = dbrvalue.get()

    # special request type
    if dbrtype == ca.DBR.CLASS_NAME:
        assert value == 'ao'
        return

    if dbrtype == ca.DBR.STSACK_STRING:
        assert value['value'] == '1.0000'
        assert value['ackt'] == 1
        assert value['acks'] == ca.AlarmSeverity.Major
        return

    value_expected = 1
    if dbrtype.isSTRING():
        value_expected = '1.0000'

    if dbrtype.isPlain():
        assert value == value_expected
    else:
        assert value['value'] == value_expected
        assert value['status'] == ca.AlarmCondition.No
        assert value['severity'] == ca.AlarmSeverity.No

        if dbrtype.isTIME():
            assert isinstance(value['stamp'], dict)
            assert 'seconds' in value['stamp']
            assert 'nanoseconds' in value['stamp']

        if dbrtype.isSTRING():
            return

        if dbrtype.isGR() or dbrtype.isCTRL():
            if dbrtype.isENUM():
                assert value['no_str'] == 0
                assert value['strs'] == ()
            else:
                lo = -10
                lolo = -20
                # char type is uint8
                if dbrtype.isCHAR():
                    lo &= 0xff
                    lolo &= 0xff
                assert value['units'] == 'mm'
                assert value['upper_disp_limit'] == 20
                assert value['lower_disp_limit'] == lolo
                assert value['upper_warning_limit'] == 10
                assert value['lower_warning_limit'] == lo
                assert value['upper_alarm_limit'] == 20
                assert value['lower_alarm_limit'] == lolo
                if dbrtype.isFLOAT() or dbrtype.isDOUBLE():
                    assert value['precision'] == 4
                if dbrtype.isCTRL():
                    assert value['upper_ctrl_limit'] == 0
                    assert value['lower_ctrl_limit'] == 0


def teardown_module(module):
    global chid
    # clear channel
    ca.clear_channel(chid)

    # destroy context
    ca.destroy_context()



"""
The macros defined macros.py are exported as IntEnum.

"""
from enum import IntEnum
from .ca import ffi, libca
from .compat import to_string
from .macros import *

__all__ = ['ChannelState', 'CA_OP', 'CA_PRIORITY',
           'CA_K', 'ECA', 'DBE',
           'AlarmSeverity', 'AlarmCondition']


class ChannelState(IntEnum):
    """
    Enum redefined from C enum channel_state
    """
    NEVER_CONN  = cs_never_conn #: valid chid, IOC not found
    PREV_CONN   = cs_prev_conn #: valid chid, IOC was found, but unavailable
    CONN        = cs_conn #: valid chid, IOC was found, still available
    CLOSED      = cs_closed #: channel deleted
    NEVER_SEARCH= cs_never_search # invalid chid


class CA_OP(IntEnum):
    """
    Enum redefined from C macros CA_OP_XXX
    """
    GET             = CA_OP_GET #: get value
    PUT             = CA_OP_PUT #: put value
    CREATE_CHANNEL  = CA_OP_CREATE_CHANNEL #: create channel
    ADD_EVENT       = CA_OP_ADD_EVENT #: subscribe
    CLEAR_EVENT     = CA_OP_CLEAR_EVENT #: unsubscribe
    OTHER           = CA_OP_OTHER #: other
    CONN_UP         = CA_OP_CONN_UP #: connection established
    CONN_DOWN       = CA_OP_CONN_DOWN #: connection lost


class CA_PRIORITY(IntEnum):
    """
    Enum redefined from CA_PRIORITY_XXX macros.
    """
    MAX     = CA_PRIORITY_MAX
    MIN     = CA_PRIORITY_MIN
    DEFAULT = CA_PRIORITY_DEFAULT

    DB_LINKS= CA_PRIORITY_DB_LINKS
    ARCHIVE = CA_PRIORITY_ARCHIVE
    OPI     = CA_PRIORITY_OPI


class DBE(IntEnum):
    """
    Enum redefined from DBE_XXX macros.
    """
    #: Trigger an event when a significant change in the channel's value
    #: occurs. Relies on the monitor deadband field under DCT.
    VALUE       = DBE_VALUE

    #: Trigger an event when an archive significant change in the channel's
    #: value occurs. Relies on the archiver monitor deadband field under DCT.
    ARCHIVE     = DBE_ARCHIVE

    #: Trigger an event when the alarm state changes
    ALARM       = DBE_ALARM

    #: Trigger an event when a property change (control limit, graphical
    #: limit, status string, enum string ...) occurs.
    PROPERTY    = DBE_PROPERTY


class CA_K(IntEnum):
    """
    Enum redefined from CA_K_XXX
    """
    INFO    = CA_K_INFO
    ERROR   = CA_K_ERROR
    SUCCESS = CA_K_SUCCESS
    WARNING = CA_K_WARNING
    SEVERE  = CA_K_SEVERE
    FATAL   = CA_K_FATAL


class ECA(IntEnum):
    """
    Enum redefined from ECA_XXX status code

    .. note:: ARRAY16KCLIENT is used in place of ECA_16KARRAYCLIENT, while variable
       name cannot start with a number.
    """
    NORMAL          = ECA_NORMAL #: Normal successful completion
    MAXIOC          = ECA_MAXIOC
    UKNHOST         = ECA_UKNHOST
    UKNSERV         = ECA_UKNSERV
    SOCK            = ECA_SOCK
    CONN            = ECA_CONN
    ALLOCMEM        = ECA_ALLOCMEM
    UKNCHAN         = ECA_UKNCHAN
    UKNFIELD        = ECA_UKNFIELD
    TOLARGE         = ECA_TOLARGE
    TIMEOUT         = ECA_TIMEOUT #: User specified timeout on IO operation expired
    NOSUPPORT       = ECA_NOSUPPORT
    STRTOBIG        = ECA_STRTOBIG
    DISCONNCHID     = ECA_DISCONNCHID #: The request was ignored because the specified channel is disconnected
    BADTYPE         = ECA_BADTYPE
    CHIDNOTFOUND    = ECA_CHIDNOTFOUND
    CHIDRETRY       = ECA_CHIDRETRY
    INTERNAL        = ECA_INTERNAL
    DBLCLFAIL       = ECA_DBLCLFAIL
    GETFAIL         = ECA_GETFAIL
    PUTFAIL         = ECA_PUTFAIL
    ADDFAIL         = ECA_ADDFAIL
    BADCOUNT        = ECA_BADCOUNT
    BADSTR          = ECA_BADSTR
    DISCONN         = ECA_DISCONN #: Virtual circuit disconnect
    DBLCHNL         = ECA_DBLCHNL
    EVDISALLOW      = ECA_EVDISALLOW
    BUILDGET        = ECA_BUILDGET
    NEEDSFP         = ECA_NEEDSFP
    OVEVFAIL        = ECA_OVEVFAIL
    BADMONID        = ECA_BADMONID
    NEWADDR         = ECA_NEWADDR
    NEWCONN         = ECA_NEWCONN
    NOCACTX         = ECA_NOCACTX
    DEFUNCT         = ECA_DEFUNCT
    EMPTYSTR        = ECA_EMPTYSTR
    NOREPEATER      = ECA_NOREPEATER
    NOCHANMSG       = ECA_NOCHANMSG
    DLCKREST        = ECA_DLCKREST
    SERVBEHIND      = ECA_SERVBEHIND
    NOCAST          = ECA_NOCAST
    BADMASK         = ECA_BADMASK
    IODONE          = ECA_IODONE
    IOINPROGRESS    = ECA_IOINPROGRESS
    BADSYNCGRP      = ECA_BADSYNCGRP
    PUTCBINPROG     = ECA_PUTCBINPROG
    NORDACCESS      = ECA_NORDACCESS
    NOWTACCESS      = ECA_NOWTACCESS
    ANACHRONISM     = ECA_ANACHRONISM
    NOSEARCHADDR    = ECA_NOSEARCHADDR
    NOCONVERT       = ECA_NOCONVERT
    BADCHID         = ECA_BADCHID
    BADFUNCPTR      = ECA_BADFUNCPTR
    ISATTACHED      = ECA_ISATTACHED
    UNAVAILINSERV   = ECA_UNAVAILINSERV
    CHANDESTROY     = ECA_CHANDESTROY
    BADPRIORITY     = ECA_BADPRIORITY
    NOTTHREADED     = ECA_NOTTHREADED
    # 16KARRAYCLIENT is an invalid variable name
    ARRAY16KCLIENT  = ECA_16KARRAYCLIENT
    CONNSEQTMO      = ECA_CONNSEQTMO
    UNRESPTMO       = ECA_UNRESPTMO

    def severity(self):
        """
        :return: the severity of the status code
        :rtype: :class:`CA_K`
        """
        return CA_K(CA_EXTRACT_MSG_SEVERITY(self.value))

    def message(self):
        """
        :return: the string representation of the status code
        """
        return to_string(ffi.string(libca.ca_message(self.value)))


class AlarmSeverity(IntEnum):
    """
    Enum redefined from C enum type epicsAlarmSeverity.
    Due to the enum difference between C and Python,
    the enum item name has been greatly simplified::

        epicsSevNone  -> AlarmSeverity.No
        epicsSevMinor -> AlarmSeverity.Minor
        ...

    .. note:: *No* is used in place of *None*, which is Python keyword.
    """
    # None is a Python keyword
    No      =  NO_ALARM
    Minor   =  MINOR_ALARM
    Major   =  MAJOR_ALARM
    Invalid =  INVALID_ALARM


class AlarmCondition(IntEnum):
    """
    Enum redefined from C enum type epicsAlarmCondition.
    Due to the enum difference between C and Python,
    the enum item name has been greatly simplified::

        epicsAlarmNone -> AlarmCondition.No
        epicsAlarmRead -> AlarmCondition.Read
        ...

    .. note:: *No* is used in place of *None*, which is Python keyword.
    """
    No         = NO_ALARM
    Read       = READ_ALARM
    Write      = WRITE_ALARM
    HiHi       = HIHI_ALARM
    High       = HIGH_ALARM
    LoLo       = LOLO_ALARM
    Low        = LOW_ALARM
    State      = STATE_ALARM
    Cos        = COS_ALARM
    Comm       = COMM_ALARM
    Timeout    = TIMEOUT_ALARM
    HwLimit    = HW_LIMIT_ALARM
    Calc       = CALC_ALARM
    Scan       = SCAN_ALARM
    Link       = LINK_ALARM
    Soft       = SOFT_ALARM
    BadSub     = BAD_SUB_ALARM
    UDF        = UDF_ALARM
    Disable    = DISABLE_ALARM
    Simm       = SIMM_ALARM
    ReadAccess = READ_ACCESS_ALARM
    WriteAccess=WRITE_ACCESS_ALARM



from __future__ import print_function
import pytest
import caffi.ca as ca

def setup_module(module):
    # create preemptive enabled context
    status = ca.create_context(True)
    assert status == ca.ECA.NORMAL


@pytest.mark.parametrize("name, input, expected", [
    ('cawaves', ['1', '2', '3', '4'], ['1', '2', '3']),
    ('cawaves', '1', ['1', '', '']),
    ('cawaveh', [1, 2, 3, 4], None),
    ('cawavef', [1, 2, 3, 4], None),
    ('cawavec', [1, 2, 3, 4], None),
    ('cawavel', [1, 2, 3, 4], None),
    ('cawave',  [1, 2, 3, 4], None),
    ('cawave',  1,            [1, 0, 0, 0]),
    ('cawave',  '1.23',       [1.23, 0, 0, 0]),
    ('cabo',    [1, 2, 3, 4], 1),
    ('cabo',    'Done',       0),
    ('catest',  1.23,         None),
    ('catest',  [1, 2, 3, 4], 1),
    ('catest',  '1.23',       1.23),
    ('cawavec', '1.23',       [49, 46, 50, 51])])
def test_put_get(name, input, expected):
    if expected is None:
        expected = input

    status, chid = ca.create_channel(name)
    assert status == ca.ECA.NORMAL

    status = ca.pend_io(10)
    assert status == ca.ECA.NORMAL

    status = ca.put(chid, input)
    assert status == ca.ECA.NORMAL

    ca.flush_io()

    status, value = ca.get(chid, count=4)
    assert status == ca.ECA.NORMAL

    status = ca.pend_io(10)
    assert status == ca.ECA.NORMAL

    assert value.get() == expected

    ca.clear_channel(chid)

def teardown_module(module):
    # destroy context
    ca.destroy_context()



from enum import IntEnum

# convert epics ca types to numpy data types
try:
    import numpy
    ctype2dtype = {
        'dbr_int_t':    numpy.int16,
        'dbr_float_t':  numpy.float32,
        'dbr_char_t':   numpy.uint8,
        'dbr_enum_t':   numpy.uint16,
        'dbr_long_t':   numpy.int32,
        'dbr_double_t': numpy.float64,
    }
    has_numpy = True
except ImportError:
    numpy = None
    has_numpy = False

from .constants import AlarmCondition, AlarmSeverity
from .compat import to_string
from .macros import *
from .ca import ffi, libca

__all__ = ['DBF', 'DBR', 'DBRValue', 'format_dbr']


#
# ptr to value given a pointer to the structure and the DBR type
#
def dbr_value_ptr(PDBR, DBR_TYPE):
    return ffi.cast('char*', PDBR) + libca.dbr_value_offset[DBR_TYPE]


class DBR(IntEnum):
    """
    Enum redefined from DBR_XXX macros.
    """
    INVALID     = TYPENOTCONN
    STRING      = DBR_STRING
    INT         = DBR_INT
    SHORT       = DBR_SHORT
    FLOAT       = DBR_FLOAT
    ENUM        = DBR_ENUM
    CHAR        = DBR_CHAR
    LONG        = DBR_LONG
    DOUBLE      = DBR_DOUBLE
    STS_STRING  = DBR_STS_STRING
    STS_SHORT   = DBR_STS_SHORT
    STS_INT     = DBR_STS_INT
    STS_FLOAT   = DBR_STS_FLOAT
    STS_ENUM    = DBR_STS_ENUM
    STS_CHAR    = DBR_STS_CHAR
    STS_LONG    = DBR_STS_LONG
    STS_DOUBLE  = DBR_STS_DOUBLE
    TIME_STRING = DBR_TIME_STRING
    TIME_INT    = DBR_TIME_INT
    TIME_SHORT  = DBR_TIME_SHORT
    TIME_FLOAT  = DBR_TIME_FLOAT
    TIME_ENUM   = DBR_TIME_ENUM
    TIME_CHAR   = DBR_TIME_CHAR
    TIME_LONG   = DBR_TIME_LONG
    TIME_DOUBLE = DBR_TIME_DOUBLE
    GR_STRING   = DBR_GR_STRING
    GR_SHORT    = DBR_GR_SHORT
    GR_INT      = DBR_GR_INT
    GR_FLOAT    = DBR_GR_FLOAT
    GR_ENUM     = DBR_GR_ENUM
    GR_CHAR     = DBR_GR_CHAR
    GR_LONG     = DBR_GR_LONG
    GR_DOUBLE   = DBR_GR_DOUBLE
    CTRL_STRING = DBR_CTRL_STRING
    CTRL_SHORT  = DBR_CTRL_SHORT
    CTRL_INT    = DBR_CTRL_INT
    CTRL_FLOAT  = DBR_CTRL_FLOAT
    CTRL_ENUM   = DBR_CTRL_ENUM
    CTRL_CHAR   = DBR_CTRL_CHAR
    CTRL_LONG   = DBR_CTRL_LONG
    CTRL_DOUBLE = DBR_CTRL_DOUBLE
    PUT_ACKT    = DBR_PUT_ACKT
    PUT_ACKS    = DBR_PUT_ACKS
    STSACK_STRING=DBR_STSACK_STRING
    CLASS_NAME  = DBR_CLASS_NAME

    def isSTRING(self):
        """
        :return: True if type is STRING or one of XXX_STRING
        """
        return dbr_type_is_STRING(self.value)

    def isSHORT(self):
        """
        :return: True if type is SHORT or one of XXX_SHORT
        """
        return dbr_type_is_SHORT(self.value)

    def isFLOAT(self):
        """
        :return: True if type is FLOAT or one of XXX_FLOAT
        """
        return dbr_type_is_FLOAT(self.value)

    def isENUM(self):
        """
        :return: True if type is ENUM or one of XXX_ENUM
        """
        return dbr_type_is_ENUM(self.value)

    def isCHAR(self):
        """
        :return: True if type is CHAR or one of XXX_CHAR
        """
        return dbr_type_is_CHAR(self.value)

    def isLONG(self):
        """
        :return: True if type is LONG or one of XXX_LONG
        """
        return dbr_type_is_LONG(self.value)

    def isDOUBLE(self):
        """
        :return: True if type is DOUBLE or one of XXX_DOUBLE
        """
        return dbr_type_is_DOUBLE(self.value)

    def isPlain(self):
        """
        :return: True if type is one of STRING, SHORT, FLOAT, ENUM, CHAR, LONG, DOUBLE
        """
        return dbr_type_is_plain(self.value)

    def isSTS(self):
        """
        :return: True if type is one of STS_XXX
        """
        return dbr_type_is_STS(self.value)

    def isTIME(self):
        """
        :return: True if type is one of TIME_XXX
        """
        return dbr_type_is_TIME(self.value)

    def isGR(self):
        """
        :return: True if type is one of GR_XXX
        """
        return dbr_type_is_GR(self.value)

    def isCTRL(self):
        """
        :return: True if type is one of CTRL_XXX
        """
        return dbr_type_is_CTRL(self.value)


class DBF(IntEnum):
    """
    Enum redefined from DBF_XXX macros.
    """
    INVALID = TYPENOTCONN #: the channel's native type when disconnected
    STRING  = DBF_STRING #: array of 40 characters
    SHORT   = DBF_SHORT #: int16
    INT     = DBF_INT #: int16, synonym of SHORT
    FLOAT   = DBF_FLOAT #: float
    ENUM    = DBF_ENUM #: uint16
    CHAR    = DBF_CHAR #: uint8
    LONG    = DBF_LONG #: int32
    DOUBLE  = DBF_DOUBLE #: double

    def toSTS(self):
        """
        :return: DBR.STS_XXX
        """
        return DBR(dbf_type_to_DBR_STS(self.value))

    def toTIME(self):
        """
        :return: DBR.TIME_XXX
        """
        return DBR(dbf_type_to_DBR_TIME(self.value))

    def toGR(self):
        """
        :return: DBR.GR_XXX
        """
        return DBR(dbf_type_to_DBR_GR(self.value))

    def toCTRL(self):
        """
        :return: DBR.CTRL_XXX
        """
        return DBR(dbf_type_to_DBR_GR(self.value))


#
# Functions that convert from DBR structure to dict
#
def format_dbr_sts(cvalue, value):
    value['status'] = AlarmCondition(cvalue.status)
    value['severity'] = AlarmSeverity(cvalue.severity)


def format_dbr_time(cvalue, value):
    secs_posix = cvalue.stamp.secPastEpoch + POSIX_TIME_AT_EPICS_EPOCH
    value['stamp'] = {
        'seconds': secs_posix,
        'nanoseconds': cvalue.stamp.nsec,
        'timestamp': secs_posix + cvalue.stamp.nsec / 1e9
    }


def format_dbr_gr(cvalue, value):
    value['units'] = to_string(ffi.string(cvalue.units))
    value['upper_disp_limit'] = cvalue.upper_disp_limit
    value['lower_disp_limit'] = cvalue.lower_disp_limit
    value['upper_alarm_limit'] = cvalue.upper_alarm_limit
    value['upper_warning_limit'] = cvalue.upper_warning_limit
    value['lower_alarm_limit'] = cvalue.lower_alarm_limit
    value['lower_warning_limit'] = cvalue.lower_warning_limit


def format_dbr_ctrl(cvalue, value):
    value['upper_ctrl_limit'] = cvalue.upper_ctrl_limit
    value['lower_ctrl_limit'] = cvalue.lower_ctrl_limit


def format_dbr_enum(cvalue, value):
    no_str = cvalue.no_str
    value['no_str'] = no_str
    value['strs'] = tuple(to_string(ffi.string(cstr)) for cstr in cvalue.strs[0:no_str])


def format_plain_value(valueType, count, cvalue, use_numpy):
    if count == 1:
        value = ffi.cast(valueType+'*', cvalue)[0]
    else:
        cvalue = ffi.cast(valueType+'[%d]'%count, cvalue)

        if has_numpy and use_numpy:
            value = numpy.frombuffer(ffi.buffer(cvalue), dtype=ctype2dtype[valueType]).copy()
        else:
            value = list(cvalue)

    return value


def format_string_value(count, dbrValue):
    cvalue = ffi.cast('dbr_string_t*', dbrValue)
    if count == 1:
        value = to_string(ffi.string(cvalue[0]))
    else:
        value = []
        for i in range(count):
            value.append(to_string(ffi.string(cvalue[i])))
    return value


def format_dbr(dbrType, count, dbrValue, use_numpy):
    """
    Convert the specified dbr data structure to Python dict

    :param dbrType: The data type, DBR_XXX
    :param count: The array element count
    :param dbrValue: A pointer of data of the specified type and number
    :return: A dict filled with the values from the C structure fields.

    """
    if dbrType == DBR_STRING:
        value = format_string_value(count, dbrValue)

    elif dbrType == DBR_INT:
        value = format_plain_value('dbr_int_t', count, dbrValue, use_numpy)

    elif dbrType == DBR_FLOAT:
        value = format_plain_value('dbr_float_t', count, dbrValue, use_numpy)

    elif dbrType == DBR_ENUM:
        value = format_plain_value('dbr_enum_t', count, dbrValue, use_numpy)

    elif dbrType == DBR_CHAR:
        value = format_plain_value('dbr_char_t', count, dbrValue, use_numpy)

    elif dbrType == DBR_LONG:
        value = format_plain_value('dbr_long_t', count, dbrValue, use_numpy)

    elif dbrType == DBR_DOUBLE:
        value = format_plain_value('dbr_double_t', count, dbrValue, use_numpy)

    elif dbrType == DBR_STS_STRING or dbrType == DBR_GR_STRING or dbrType == DBR_CTRL_STRING:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_string*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_string_value(count, dbr_value_ptr(cvalue, dbrType))

    elif dbrType == DBR_STS_INT:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_int*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_plain_value('dbr_int_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_STS_FLOAT:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_float*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_plain_value('dbr_float_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_STS_ENUM:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_enum*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_plain_value('dbr_enum_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_STS_CHAR:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_char*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_plain_value('dbr_char_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_STS_LONG:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_long*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_plain_value('dbr_long_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_STS_DOUBLE:
        value = {}
        cvalue = ffi.cast('struct dbr_sts_double*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['value'] = format_plain_value('dbr_double_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_TIME_STRING:
        value = {}
        cvalue = ffi.cast('struct dbr_time_string*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_string_value(count, dbr_value_ptr(cvalue, dbrType))

    elif dbrType == DBR_TIME_INT:
        value = {}
        cvalue = ffi.cast('struct dbr_time_int*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_plain_value('dbr_int_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_TIME_FLOAT:
        value = {}
        cvalue = ffi.cast('struct dbr_time_float*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_plain_value('dbr_float_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_TIME_ENUM:
        value = {}
        cvalue = ffi.cast('struct dbr_time_enum*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_plain_value('dbr_enum_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_TIME_CHAR:
        value = {}
        cvalue = ffi.cast('struct dbr_time_char*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_plain_value('dbr_char_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_TIME_LONG:
        value = {}
        cvalue = ffi.cast('struct dbr_time_long*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_plain_value('dbr_long_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_TIME_DOUBLE:
        value = {}
        cvalue = ffi.cast('struct dbr_time_double*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_time(cvalue, value)
        value['value'] = format_plain_value('dbr_double_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_GR_INT:
        value = {}
        cvalue = ffi.cast('struct dbr_gr_int*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['value'] = format_plain_value('dbr_int_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_GR_FLOAT:
        value = {}
        cvalue = ffi.cast('struct dbr_gr_float*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['precision'] = cvalue.precision
        value['value'] = format_plain_value('dbr_float_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_GR_ENUM:
        value = {}
        cvalue = ffi.cast('struct dbr_gr_enum*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_enum(cvalue, value)
        value['value'] = format_plain_value('dbr_enum_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_GR_CHAR:
        value = {}
        cvalue = ffi.cast('struct dbr_gr_char*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['value'] = format_plain_value('dbr_char_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_GR_LONG:
        value = {}
        cvalue = ffi.cast('struct dbr_gr_long*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['value'] = format_plain_value('dbr_long_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_GR_DOUBLE:
        value = {}
        cvalue = ffi.cast('struct dbr_gr_double*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['precision'] = cvalue.precision
        value['value'] = format_plain_value('dbr_double_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CTRL_INT:
        value = {}
        cvalue = ffi.cast('struct dbr_ctrl_int*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        format_dbr_ctrl(cvalue, value)
        value['value'] = format_plain_value('dbr_int_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CTRL_FLOAT:
        value = {}
        cvalue = ffi.cast('struct dbr_ctrl_float*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['precision'] = cvalue.precision
        format_dbr_ctrl(cvalue, value)
        value['value'] = format_plain_value('dbr_float_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CTRL_ENUM:
        value = {}
        cvalue = ffi.cast('struct dbr_ctrl_enum*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_enum(cvalue, value)
        value['value'] = format_plain_value('dbr_enum_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CTRL_CHAR:
        value = {}
        cvalue = ffi.cast('struct dbr_ctrl_char*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        format_dbr_ctrl(cvalue, value)
        value['value'] = format_plain_value('dbr_char_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CTRL_LONG:
        value = {}
        cvalue = ffi.cast('struct dbr_ctrl_long*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        format_dbr_ctrl(cvalue, value)
        value['value'] = format_plain_value('dbr_long_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CTRL_DOUBLE:
        value = {}
        cvalue = ffi.cast('struct dbr_ctrl_double*', dbrValue)
        format_dbr_sts(cvalue, value)
        format_dbr_gr(cvalue, value)
        value['precision'] = cvalue.precision
        format_dbr_ctrl(cvalue, value)
        value['value'] = format_plain_value('dbr_double_t', count, dbr_value_ptr(cvalue, dbrType), use_numpy)

    elif dbrType == DBR_CLASS_NAME:
        value = format_string_value(count, dbrValue)

    elif dbrType == DBR_STSACK_STRING:
        value = {}
        cvalue = ffi.cast('struct dbr_stsack_string*', dbrValue)
        format_dbr_sts(cvalue, value)
        value['ackt'] = cvalue.ackt
        value['acks'] = AlarmSeverity(cvalue.acks)
        value['value'] = format_string_value(count, dbr_value_ptr(cvalue, dbrType))

    # It should never reach here
    else:
        value = None

    return value


class DBRValue(object):
    """
    :param dbrtype: The external type of the supplied *cvalue*
    :param count: Element count of the supplied *cvalue*
    :param cvalue: Pointer to the structure of *dbrtype* with *count* element
    :param bool use_numpy: whether to format numeric waveform as numpy array

    An convenient object to represent the value returned by :func:`caffi.ca.get` and :func:`caffi.ca.sg_get`.
    It holds the reference to the memory allocated by the get functions,
    in addition the type and element count information.

    Once the memory is assured to be stable, normally when the gets function completed with success,
    call :meth:`get` to get the returned values.

    """
    def __init__(self, dbrtype=DBR.INVALID, count=0, cvalue=ffi.NULL, use_numpy=False):
        """
        """
        self.dbrtype = dbrtype
        self.count = count
        self.cvalue = cvalue
        self.use_numpy = use_numpy

    def get(self):
        """
        :return: Value for plain DBR_XXXX type or a dict for DBR_STS_XXXX etc.

        .. note:: This method should be called only if the get request has succeeded.
        """
        return format_dbr(self.dbrtype, self.count, self.cvalue, self.use_numpy)



"""
Compatible layer for Python 2 and 3
"""
import sys

__all__ = ['basestring', 'to_bytes', 'to_string']

if sys.hexversion >= 0x03000000:
    basestring = (str, bytes)
else:
    basestring = str


def to_bytes(str_val):
    if isinstance(str_val, bytes):
        return str_val

    if sys.hexversion >= 0x03000000:
        return str_val.encode('utf8')
    else:
        return str(str_val)


def to_string(bytes_val):
    if isinstance(bytes_val, str):
        return bytes_val

    if sys.hexversion >= 0x03000000:
        return bytes_val.decode('utf8')
    else:
        return str(bytes_val)



from __future__ import (print_function, absolute_import)
import threading
from pprint import pprint
import caffi.ca as ca


def setup_module(module):
    # create context
    status = ca.create_context(True)
    assert status == ca.ECA.NORMAL


def teardown_module(module):
    # destroy context
    ca.destroy_context()


def exceptionCB(epicsArgs):
    pprint('exception callback')
    pprint(epicsArgs, indent=4)


def connectCB(epicsArgs, event):
    pprint('connect callback')
    pprint(epicsArgs, indent=4)
    event.set()


def putCB(epicsArgs, event):
    pprint('put callback')
    pprint(epicsArgs, indent=4)
    event.set()


def getCB(epicsArgs, event):
    pprint('get callback')
    pprint(epicsArgs, indent=4)
    event.set()


def monitorCB(epicsArgs, event):
    pprint('monitor callback')
    pprint(epicsArgs, indent=4)
    event.set()


def accessCB(epicsArgs, event):
    pprint('access rights callback')
    pprint(epicsArgs, indent=4)
    event.set()


def test_callbacks():
    # events to notify callback has be called
    connect_done = threading.Event()
    get_done = threading.Event()
    put_done = threading.Event()
    access_done = threading.Event()
    monitor_done = threading.Event()

    # exception callback
    status = ca.add_exception_event(exceptionCB)
    assert status == ca.ECA.NORMAL

    # connection callback
    status, chid = ca.create_channel('cawave', callback=lambda arg: connectCB(arg, connect_done))
    assert status == ca.ECA.NORMAL

    # access rights callback
    status = ca.replace_access_rights_event(chid, lambda arg: accessCB(arg, access_done))
    assert status == ca.ECA.NORMAL

    # since preemptive context is used, the background activities happen asynchronously
    connect_done.wait(5)
    assert ca.state(chid) == ca.ChannelState.CONN

    # access callback is called whenever connection changed
    access_done.wait(5)
    assert ca.read_access(chid)
    assert ca.write_access(chid)

    # monitor callback
    status, evid = ca.create_subscription(chid, callback=lambda arg: monitorCB(arg, monitor_done), count=3)
    assert status == ca.ECA.NORMAL

    # push request to server
    ca.flush_io()

    # access callback is called once it is registered
    monitor_done.wait(5)

    # put callback
    status = ca.put(chid, [60, 50, 40], callback=lambda arg: putCB(arg, put_done))
    assert status == ca.ECA.NORMAL

    # push these requests to server
    ca.flush_io()

    # wait for put callback
    put_done.wait(5)

    # get callback, the dbrvalue is None since callback is used
    status, dbrvalue = ca.get(chid, callback=lambda arg: getCB(arg, get_done), count=3)
    assert status == ca.ECA.NORMAL

    ca.flush_io()

    # wait for callbacks to happen
    get_done.wait(5)

    # exception callback
    status = ca.add_exception_event()
    assert status == ca.ECA.NORMAL

    # clean up
    ca.clear_subscription(evid)
    ca.clear_channel(chid)
    ca.flush_io()



from __future__ import print_function
import datetime
import threading
import time
import caffi.ca as ca

def setup_module(module):
    # create explicitly a preemptive enabled context
    # so that it can attached in other threads
    status = ca.create_context(True)
    assert status == ca.ECA.NORMAL

    global ctx
    ctx = ca.current_context()

    # create channel
    global chid
    status, chid = ca.create_channel('catest')
    assert status == ca.ECA.NORMAL

    # wait for connections
    status = ca.pend_io(3)
    assert status == ca.ECA.NORMAL

    # put
    status = ca.put(chid, 0)
    assert status == ca.ECA.NORMAL

    ca.flush_io()

    def monitor(epics_arg):
        print(datetime.datetime.fromtimestamp(epics_arg['value']['stamp']['timestamp']), epics_arg['value']['value'])

    global evid
    status, evid = ca.create_subscription(chid, monitor, chtype=ca.DBR.TIME_DOUBLE)


def thread_even():
    status = ca.attach_context(ctx)
    assert status == ca.ECA.NORMAL

    for i in range(0, 200, 2):
        ca.put(chid,  i)
        ca.flush_io()


def thread_odd():
    status = ca.attach_context(ctx)
    assert status == ca.ECA.NORMAL

    for i in range(1, 200, 2):
        ca.put(chid, i)
        ca.flush_io()


def test_thread():
    tid1 = threading.Thread(target=thread_odd)
    tid2 = threading.Thread(target=thread_even)
    tid1.start()
    tid2.start()
    tid1.join()
    tid2.join()

    # wait for the last monitor to arrive
    time.sleep(2)

    # the value should be either 199 or 198, depending on which thread ends last
    status, value = ca.get(chid)
    assert status == ca.ECA.NORMAL

    status = ca.pend_io(3)
    assert status == ca.ECA.NORMAL

    assert value.get() in [198, 199]


def teardown_module(module):
    # clear subscription
    ca.clear_subscription(evid)

    # clear channel
    ca.clear_channel(chid)

    ca.flush_io()

    # destroy context
    ca.destroy_context()



"""
The macros defined in C header files, *cadef.h*, *caeventmask.h*, *caerr.h*

"""
from ._ca import libca as _libca
from ._ca import ffi as _ffi

from . import compat as _compat

cs_never_conn = _libca.cs_never_conn
cs_prev_conn = _libca.cs_prev_conn
cs_conn = _libca.cs_conn
cs_closed = _libca.cs_closed
cs_never_search = 4

CA_OP_GET = _libca.CA_OP_GET
CA_OP_PUT = _libca.CA_OP_PUT
CA_OP_CREATE_CHANNEL = _libca.CA_OP_CREATE_CHANNEL
CA_OP_ADD_EVENT = _libca.CA_OP_ADD_EVENT
CA_OP_CLEAR_EVENT = _libca.CA_OP_CLEAR_EVENT
CA_OP_OTHER = _libca.CA_OP_OTHER
CA_OP_CONN_UP = _libca.CA_OP_CONN_UP
CA_OP_CONN_DOWN = _libca.CA_OP_CONN_DOWN

CA_PRIORITY_MAX = _libca.CA_PRIORITY_MAX
CA_PRIORITY_MIN = _libca.CA_PRIORITY_MIN
CA_PRIORITY_DEFAULT = _libca.CA_PRIORITY_DEFAULT

CA_PRIORITY_DB_LINKS = _libca.CA_PRIORITY_DB_LINKS
CA_PRIORITY_ARCHIVE = _libca.CA_PRIORITY_ARCHIVE
CA_PRIORITY_OPI = _libca.CA_PRIORITY_OPI

DBE_VALUE = _libca.DBE_VALUE
DBE_ARCHIVE = _libca.DBE_ARCHIVE
DBE_LOG = _libca.DBE_LOG
DBE_ALARM = _libca.DBE_ALARM
DBE_PROPERTY = _libca.DBE_PROPERTY

# db_access.h
TYPENOTCONN = -1 #/* the channel's native type when disconnected   */

# /* database field types */
DBF_STRING  = 0
DBF_INT     = 1
DBF_SHORT   = 1
DBF_FLOAT   = 2
DBF_ENUM    = 3
DBF_CHAR    = 4
DBF_LONG    = 5
DBF_DOUBLE  = 6
DBF_NO_ACCESS=7
LAST_TYPE   = DBF_DOUBLE

# /* data request buffer types */
DBR_STRING      = DBF_STRING
DBR_INT	        = DBF_INT
DBR_SHORT       = DBF_INT
DBR_FLOAT       = DBF_FLOAT
DBR_ENUM        = DBF_ENUM
DBR_CHAR        = DBF_CHAR
DBR_LONG        = DBF_LONG
DBR_DOUBLE      = DBF_DOUBLE
DBR_STS_STRING  = 7
DBR_STS_SHORT   = 8
DBR_STS_INT     = DBR_STS_SHORT
DBR_STS_FLOAT   = 9
DBR_STS_ENUM    = 10
DBR_STS_CHAR    = 11
DBR_STS_LONG    = 12
DBR_STS_DOUBLE  = 13
DBR_TIME_STRING = 14
DBR_TIME_INT    = 15
DBR_TIME_SHORT  = 15
DBR_TIME_FLOAT  = 16
DBR_TIME_ENUM   = 17
DBR_TIME_CHAR   = 18
DBR_TIME_LONG   = 19
DBR_TIME_DOUBLE = 20
DBR_GR_STRING   = 21
DBR_GR_SHORT    = 22
DBR_GR_INT      = DBR_GR_SHORT
DBR_GR_FLOAT    = 23
DBR_GR_ENUM     = 24
DBR_GR_CHAR     = 25
DBR_GR_LONG     = 26
DBR_GR_DOUBLE   = 27
DBR_CTRL_STRING = 28
DBR_CTRL_SHORT  = 29
DBR_CTRL_INT    = DBR_CTRL_SHORT
DBR_CTRL_FLOAT  = 30
DBR_CTRL_ENUM   = 31
DBR_CTRL_CHAR   = 32
DBR_CTRL_LONG   = 33
DBR_CTRL_DOUBLE = 34
DBR_PUT_ACKT    = DBR_CTRL_DOUBLE + 1
DBR_PUT_ACKS    = DBR_PUT_ACKT + 1
DBR_STSACK_STRING=DBR_PUT_ACKS + 1
DBR_CLASS_NAME  = DBR_STSACK_STRING + 1
LAST_BUFFER_TYPE= DBR_CLASS_NAME

#/*
# * type checking macros -- return non-zero if condition is true, zero otherwise
# */
def dbf_type_is_valid(dbftype):
    return dbftype >= 0 and dbftype <= LAST_TYPE


def dbr_type_is_valid(dbrtype):
    return dbrtype >= 0 and dbrtype <= LAST_BUFFER_TYPE


def dbr_type_is_plain(dbrtype):
    return dbrtype >= DBR_STRING and dbrtype <= DBR_DOUBLE


def dbr_type_is_STS(dbrtype):
    return dbrtype >= DBR_STS_STRING and dbrtype <= DBR_STS_DOUBLE


def dbr_type_is_TIME(dbrtype):
    return dbrtype >= DBR_TIME_STRING and dbrtype <= DBR_TIME_DOUBLE


def dbr_type_is_GR(dbrtype):
    return dbrtype >= DBR_GR_STRING and dbrtype <= DBR_GR_DOUBLE


def dbr_type_is_CTRL(dbrtype):
    return dbrtype >= DBR_CTRL_STRING and dbrtype <= DBR_CTRL_DOUBLE


def dbr_type_is_STRING(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_STRING )


def dbr_type_is_SHORT(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_SHORT )


def dbr_type_is_FLOAT(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_FLOAT )

def dbr_type_is_ENUM(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_ENUM )


def dbr_type_is_CHAR(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_CHAR )


def dbr_type_is_LONG(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_LONG )


def dbr_type_is_DOUBLE(dbrtype):
    return ( dbr_type_is_valid(dbrtype) and
             dbrtype % (LAST_TYPE + 1) == DBR_DOUBLE )

#/*
# * DBR structure size in bytes
# */
def dbr_size_n(TYPE,COUNT):
    if COUNT <=0:
        return _libca.dbr_size[TYPE]
    else:
        return _libca.dbr_size[TYPE] + (COUNT-1) * _libca.dbr_value_size[TYPE]

#/*
# * type conversion macros
# */
def dbf_type_to_DBR(dbftype):
    if dbf_type_is_valid(dbftype):
        return dbftype
    else:
        return -1


def dbf_type_to_DBR_STS(dbftype):
    if dbf_type_is_valid(dbftype):
        return dbftype + LAST_TYPE + 1
    else:
        return -1


def dbf_type_to_DBR_TIME(dbftype):
    if dbf_type_is_valid(dbftype):
        return dbftype + 2 * (LAST_TYPE + 1)
    else:
        return -1


def dbf_type_to_DBR_GR(dbftype):
    if dbf_type_is_valid(dbftype):
        return dbftype + 3 * (LAST_TYPE + 1)
    else:
        return -1


def dbf_type_to_DBR_CTRL(dbftype):
    if dbf_type_is_valid(dbftype):
        return dbftype + 4 * (LAST_TYPE + 1)
    else:
        return -1

#/*
# * type conversion macros
# */
def dbf_type_to_text(dbftype):
    if dbftype >= -1 and dbftype <= LAST_TYPE + 1:
        text = _ffi.string(_libca.dbf_text[dbftype + 1])
    else:
        text = _ffi.string(_libca.dbf_text_invalid)

    return _compat.to_string(text)
# alias
dbf_text = dbf_type_to_text

def dbf_text_to_type(text):
    for dbftype in range(0, LAST_TYPE + 2):
        if _compat.to_bytes(text) == _ffi.string(_libca.dbf_text[dbftype+1]):
            break
    else:
        dbftype = -1

    return dbftype

def dbr_type_to_text(dbrtype):
    if dbrtype >= 0 and dbrtype <= LAST_BUFFER_TYPE:
        text = _ffi.string(_libca.dbr_text[dbrtype])
    else:
        text = _ffi.string(_libca.dbr_text_invalid)

    return _compat.to_string(text)
# alias
dbr_text = dbr_type_to_text

def dbr_text_to_type(text):
    for dbrtype in range(0, LAST_BUFFER_TYPE + 1):
        if _compat.to_bytes(text) == _ffi.string(_libca.dbr_text[dbrtype]):
            break
    else:
        dbrtype = -1
    return dbrtype

# caerr.h
#/*  CA Status Code Definitions   */

CA_K_INFO       = 3   #/* successful */
CA_K_ERROR      = 2   #/* failed- continue */
CA_K_SUCCESS    = 1   #/* successful */
CA_K_WARNING    = 0   #/* unsuccessful */
CA_K_SEVERE     = 4   #/* failed- quit */
CA_K_FATAL      = CA_K_ERROR | CA_K_SEVERE

CA_M_MSG_NO     = 0x0000FFF8
CA_M_SEVERITY   = 0x00000007
CA_M_LEVEL      = 0x00000003
CA_M_SUCCESS    = 0x00000001
CA_M_ERROR      = 0x00000002
CA_M_SEVERE     = 0x00000004

CA_V_MSG_NO     = 0x03
CA_V_SEVERITY   = 0x00
CA_V_SUCCESS    = 0x00


def CA_EXTRACT_MSG_NO(code):
    return (code & CA_M_MSG_NO) >> CA_V_MSG_NO


def CA_EXTRACT_MSG_SEVERITY(code):
    return (code & CA_M_SEVERITY) >> CA_V_SEVERITY

ECA_NORMAL          =   1
ECA_MAXIOC          =  10
ECA_UKNHOST         =  18
ECA_UKNSERV         =  26
ECA_SOCK            =  34
ECA_CONN            =  40
ECA_ALLOCMEM        =  48
ECA_UKNCHAN         =  56
ECA_UKNFIELD        =  64
ECA_TOLARGE         =  72
ECA_TIMEOUT         =  80
ECA_NOSUPPORT       =  88
ECA_STRTOBIG        =  96
ECA_DISCONNCHID     = 106
ECA_BADTYPE         = 114
ECA_CHIDNOTFOUND    = 123
ECA_CHIDRETRY       = 131
ECA_INTERNAL        = 142
ECA_DBLCLFAIL       = 144
ECA_GETFAIL         = 152
ECA_PUTFAIL         = 160
ECA_ADDFAIL         = 168
ECA_BADCOUNT        = 176
ECA_BADSTR          = 186
ECA_DISCONN         = 192
ECA_DBLCHNL         = 200
ECA_EVDISALLOW      = 210
ECA_BUILDGET        = 216
ECA_NEEDSFP         = 224
ECA_OVEVFAIL        = 232
ECA_BADMONID        = 242
ECA_NEWADDR         = 248
ECA_NEWCONN         = 259
ECA_NOCACTX         = 264
ECA_DEFUNCT         = 278
ECA_EMPTYSTR        = 280
ECA_NOREPEATER      = 288
ECA_NOCHANMSG       = 296
ECA_DLCKREST        = 304
ECA_SERVBEHIND      = 312
ECA_NOCAST          = 320
ECA_BADMASK         = 330
ECA_IODONE          = 339
ECA_IOINPROGRESS    = 347
ECA_BADSYNCGRP      = 354
ECA_PUTCBINPROG     = 362
ECA_NORDACCESS      = 368
ECA_NOWTACCESS      = 376
ECA_ANACHRONISM     = 386
ECA_NOSEARCHADDR    = 392
ECA_NOCONVERT       = 400
ECA_BADCHID         = 410
ECA_BADFUNCPTR      = 418
ECA_ISATTACHED      = 424
ECA_UNAVAILINSERV   = 432
ECA_CHANDESTROY     = 440
ECA_BADPRIORITY     = 450
ECA_NOTTHREADED     = 458
ECA_16KARRAYCLIENT  = 464
ECA_CONNSEQTMO      = 472
ECA_UNRESPTMO       = 480


# epicsTime.h
#/* The EPICS Epoch is 00:00:00 Jan 1, 1990 UTC */
POSIX_TIME_AT_EPICS_EPOCH = 631152000

# alarm.h
NO_ALARM = _libca.epicsSevNone
MINOR_ALARM = _libca.epicsSevMinor
MAJOR_ALARM = _libca.epicsSevMajor
INVALID_ALARM = _libca.epicsSevInvalid

READ_ALARM = _libca.epicsAlarmRead
WRITE_ALARM = _libca.epicsAlarmWrite
HIHI_ALARM = _libca.epicsAlarmHiHi
HIGH_ALARM = _libca.epicsAlarmHigh
LOLO_ALARM = _libca.epicsAlarmLoLo
LOW_ALARM  = _libca.epicsAlarmLow
STATE_ALARM = _libca.epicsAlarmState
COS_ALARM = _libca.epicsAlarmCos
COMM_ALARM = _libca.epicsAlarmComm
TIMEOUT_ALARM = _libca.epicsAlarmTimeout
HW_LIMIT_ALARM = _libca.epicsAlarmHwLimit
CALC_ALARM = _libca.epicsAlarmCalc
SCAN_ALARM = _libca.epicsAlarmScan
LINK_ALARM = _libca.epicsAlarmLink
SOFT_ALARM  = _libca.epicsAlarmSoft
BAD_SUB_ALARM = _libca.epicsAlarmBadSub
UDF_ALARM = _libca.epicsAlarmUDF
DISABLE_ALARM = _libca.epicsAlarmDisable
SIMM_ALARM = _libca.epicsAlarmSimm
READ_ACCESS_ALARM = _libca.epicsAlarmReadAccess
WRITE_ACCESS_ALARM = _libca.epicsAlarmWriteAccess

def epicsAlarmSeverityStrings(severity):
    return ["NO_ALARM",
            "MINOR",
            "MAJOR",
            "INVALID"][severity]

alarmSeverityString = epicsAlarmSeverityStrings

def epicsAlarmConditionStrings(status):
    return ["NO_ALARM",
            "READ",
            "WRITE",
            "HIHI",
            "HIGH",
            "LOLO",
            "LOW",
            "STATE",
            "COS",
            "COMM",
            "TIMEOUT",
            "HWLIMIT",
            "CALC",
            "SCAN",
            "LINK",
            "SOFT",
            "BAD_SUB",
            "UDF",
            "DISABLE",
            "SIMM",
            "READ_ACCESS",
            "WRITE_ACCESS"][status]

alarmStatusString = epicsAlarmConditionStrings